import copy
import datetime
import math
from dataclasses import dataclass, field
from typing import Union, get_args

import numpy as np

__all__ = [
    "Metadata",
    "MetadataValue",
    "common_metadata",
    "metadata_diff",
]


METADATA_COMMON_ATTRIBUTES = [
    "risk_basis",
    "country",
    "currency",
    "reinsurance_basis",
    "loss_definition",
    "per_occurrence_limit",
    "details",
    "loss_details",
]


MetadataValue = Union[str, float, int, bool, datetime.date, np.ndarray, None]


@dataclass(frozen=True, eq=True)
class Metadata(object):
    # noinspection PyUnresolvedReferences
    """Metadata about the information stored in `Cell`s.

    Attributes:
        risk_basis: The basis on which risks are assigned to experience periods.
            Common values are "Accident" and "Policy".
        country: The country where the risks are domiciled, as an ISO 3166 alpha-2 code
            (e.g., "US", "ES", "DE").
        currency: The currency in which premium and losses are denominated, as an ISO 4217
            code (e.g., "USD", "GBP", "JPY").
        reinsurance_basis: Whether the losses are net of reinsurance.
            Typical values are "Gross" and "Net"; the field is typed as a `str` and not a
            `bool` to allow for more nuanced gradations of value.
        loss_definition: The portions of loss costs and expenses that are included in
            loss amounts. Typical values are "Loss", "Loss+DCC", and "Loss+LAE".
        per_occurrence_limit: The per-occurrence limit on individual losses.
        details: Other information about the data represented in the `Cell`. For example,
            `details` could plausibly include coverage, rated state, or product code.
        loss_details: Information about the data in the `Cell` that only pertains to the losses.
            For example, you may have a triangle where losses are segmented by cause of loss,
            but premium is not.
    """

    risk_basis: str = "Accident"
    country: str | None = None
    currency: str | None = None
    reinsurance_basis: str | None = None
    loss_definition: str | None = None
    per_occurrence_limit: float | None = None
    details: dict[str, MetadataValue] = field(default_factory=dict)
    loss_details: dict[str, MetadataValue] = field(default_factory=dict)

    def __post_init__(self):
        # Perform validations on string attributes
        for attr in [
            "risk_basis",
            "country",
            "currency",
            "reinsurance_basis",
            "loss_definition",
        ]:
            if not isinstance(getattr(self, attr), (str, type(None))):
                raise TypeError(f"{attr} should be a string.")

        # Perform validation on per_occurrence_limit
        if not isinstance(self.per_occurrence_limit, (int, float, type(None))):
            raise TypeError("per_occurrence_limit should be a number.")

        # Perform validations on details and loss_details objects
        for attr in ["details", "loss_details"]:
            test_attr = getattr(self, attr)

            # Test type
            if not isinstance(test_attr, dict):
                raise TypeError("{attr} should be of type dict.")

            # Test dict structure
            for k, v in test_attr.items():
                if not isinstance(k, str):
                    raise TypeError(f"{attr} has non-string dict key.")
                if not isinstance(v, get_args(MetadataValue)):
                    raise TypeError(f"{attr} has dict value of invalid type.")

    def as_dict(self):
        return {
            "currency": self.currency,
            "country": self.country,
            "risk_basis": self.risk_basis,
            "reinsurance_basis": self.reinsurance_basis,
            "loss_definition": self.loss_definition,
            "per_occurrence_limit": self.per_occurrence_limit,
            "details": copy.deepcopy(self.details),
            "loss_details": copy.deepcopy(self.loss_details),
        }

    def as_flat_dict(self):
        """All metadata associated with the cell (currency, country, details, etc.)"""
        return {
            "currency": self.currency,
            "country": self.country,
            "risk_basis": self.risk_basis,
            "reinsurance_basis": self.reinsurance_basis,
            "loss_definition": self.loss_definition,
            "per_occurrence_limit": self.per_occurrence_limit,
            **self.details,
            **self.loss_details,
        }

    def __hash__(self):
        return hash(
            (
                self.risk_basis,
                self.country,
                self.currency,
                self.reinsurance_basis,
                self.loss_definition,
                self.per_occurrence_limit,
                frozenset(self.details.items()),
                frozenset(self.loss_details.items()),
            )
        )

    def __lt__(self, other):
        """Required so metadata can be sorted. This allows cells sharing metadata to be grouped
        together when sorting lists of cells.
        Comparison is based on fields in the canonical order
        """
        # noinspection DuplicatedCode
        return (
            "" if self.risk_basis is None else self.risk_basis,
            "" if self.country is None else self.country,
            "" if self.currency is None else self.currency,
            "" if self.reinsurance_basis is None else self.reinsurance_basis,
            "" if self.loss_definition is None else self.loss_definition,
            (
                math.inf
                if self.per_occurrence_limit is None
                else self.per_occurrence_limit
            ),
            tuple(sorted(self.details.items())),
            tuple(sorted(self.loss_details.items())),
        ) < (
            "" if other.risk_basis is None else other.risk_basis,
            "" if other.country is None else other.country,
            "" if other.currency is None else other.currency,
            "" if other.reinsurance_basis is None else other.reinsurance_basis,
            "" if other.loss_definition is None else other.loss_definition,
            (
                math.inf
                if other.per_occurrence_limit is None
                else other.per_occurrence_limit
            ),
            tuple(sorted(other.details.items())),
            tuple(sorted(self.loss_details.items())),
        )


def common_metadata(meta1: Metadata, meta2: Metadata) -> Metadata:
    """Return a Metadata object that has all elements that are identical in meta1 and meta2."""
    risk_basis = _first_if_equal(meta1.risk_basis, meta2.risk_basis)
    country = _first_if_equal(meta1.country, meta2.country)
    currency = _first_if_equal(meta1.currency, meta2.currency)
    reinsurance_basis = _first_if_equal(
        meta1.reinsurance_basis, meta2.reinsurance_basis
    )
    loss_definition = _first_if_equal(meta1.loss_definition, meta2.loss_definition)
    per_occurrence_limit = _first_if_equal(
        meta1.per_occurrence_limit, meta2.per_occurrence_limit
    )

    common_detail_keys = set(meta1.details.keys()) & set(meta2.details.keys())
    details = {
        k: meta1.details[k]
        for k in common_detail_keys
        if meta1.details[k] == meta2.details[k]
    }
    common_loss_detail_keys = set(meta1.loss_details.keys()) & set(
        meta2.loss_details.keys()
    )
    loss_details = {
        k: meta2.loss_details[k]
        for k in common_loss_detail_keys
        if meta1.loss_details[k] == meta2.loss_details[k]
    }

    return Metadata(
        risk_basis=risk_basis,
        country=country,
        currency=currency,
        reinsurance_basis=reinsurance_basis,
        loss_definition=loss_definition,
        per_occurrence_limit=per_occurrence_limit,
        details=details,
        loss_details=loss_details,
    )


def _first_if_equal(arg1, arg2):
    """Return arg1 if arg1 == arg2, otherise None."""
    if arg1 == arg2:
        return arg1
    else:
        return None


def metadata_diff(meta_core: Metadata, meta_diff: Metadata) -> Metadata:
    """Return a Metadata object with all of the items in `meta_diff` that are not in `meta_core`."""

    risk_basis = meta_diff.risk_basis if meta_core.risk_basis is None else None
    country = meta_diff.country if meta_core.country is None else None
    currency = meta_diff.currency if meta_core.currency is None else None
    reinsurance_basis = (
        meta_diff.reinsurance_basis if meta_core.reinsurance_basis is None else None
    )
    loss_definition = (
        meta_diff.loss_definition if meta_core.loss_definition is None else None
    )
    per_occurrence_limit = (
        meta_diff.per_occurrence_limit
        if meta_core.per_occurrence_limit is None
        else None
    )
    details = {k: v for k, v in meta_diff.details.items() if k not in meta_core.details}
    loss_details = {
        k: v
        for k, v in meta_diff.loss_details.items()
        if k not in meta_core.loss_details
    }

    return Metadata(
        risk_basis=risk_basis,
        country=country,
        currency=currency,
        reinsurance_basis=reinsurance_basis,
        loss_definition=loss_definition,
        per_occurrence_limit=per_occurrence_limit,
        details=details,
        loss_details=loss_details,
    )



from typing import Tuple

import numpy as np

__all__ = [
    "program_earned_premium",
]


def program_earned_premium(
    premium_volume: float,
    writing_pattern: np.ndarray,
    writing_resolution: int,
    earning_pattern: np.ndarray,
    earning_resolution: int,
    output_resolution: int,
    output_offset: int = 0,
    continuous_writing: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Get the aggregate premium earning and writing pattern for a treaty period, based
    on the expected pattern of written premium.

    Args:
        premium_volume: Total premium to be written into the program (in dollars).
        writing_pattern: An array with relative amounts of premium that will be written
            at each time step. For examples, np.array([1, 2, 3, 4]) would represent an
            increasing amount of premium written at each successive time step.
        writing_resolution: The number of months in each time step in the
            `writing_pattern` argument.
        earning_pattern: An array with relative amounts of premium that will be earned
            over the duration of a policy. The length of the policy is equal to
            `len(earning_pattern) * earning_resolution`.
        earning_resolution: The number of months in each time step in the
            `earning_pattern` argument.
        output_resolution: The number of months in each time step of the output
            aggregate earning pattern.
        output_offset: The number of months between the start of the first period of
            the output earning pattern and the effective date of the first written
            policy. For example, if `output_resolution` is `3` (implying a quarterly
            time series with calendar-quarter periods) and the program is effective on
            March 1st, then `output_offset` would be `2`.
        continuous_writing: If `True`, then it is assumed that policy effective dates
            are spread evenly throughout each time period in `writing_pattern`.
            Otherwise, it is assumed that all policy effective dates are on the first
            day of each time period in `writing_pattern`.
    """
    normalized_writing = writing_pattern / np.sum(writing_pattern)
    monthly_writing = np.repeat(
        premium_volume * normalized_writing / writing_resolution,
        writing_resolution,
    )
    normalized_earning = earning_pattern / np.sum(earning_pattern)
    raw_monthly_earning = np.repeat(
        normalized_earning / earning_resolution, earning_resolution
    )
    if continuous_writing:
        monthly_earning = np.sum(
            [
                np.concatenate([raw_monthly_earning / 2, [0]]),
                np.concatenate([[0], raw_monthly_earning / 2]),
            ],
            axis=0,
        )
    else:
        monthly_earning = raw_monthly_earning

    N = len(monthly_writing)
    monthly_combined = np.sum(
        [
            written_premium
            * np.concatenate([[0] * n, monthly_earning, [0] * (N - n - 1)])
            for n, written_premium in enumerate(monthly_writing)
        ],
        axis=0,
    )
    output_earning_pattern = [0.0]
    output_writing_pattern = [0.0]

    start = 0
    stop = output_offset if output_offset > 0 else output_resolution
    while start < monthly_combined.size:
        output_writing_pattern.append(np.sum(monthly_writing[start:stop]))
        output_earning_pattern.append(np.sum(monthly_combined[start:stop]))
        start, stop = stop, stop + output_resolution

    return np.array(output_writing_pattern), np.array(output_earning_pattern)



import datetime
import gzip
import io
import math
import struct
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile

import awswrangler as wr
import boto3
import numpy as np
from botocore.config import Config

from ..base import Cell, CumulativeCell, IncrementalCell, Metadata, MetadataValue
from ..triangle import Triangle
from .binary import *

__all__ = [
    "binary_to_triangle",
]


# See binary_output.py for more extensive comments on the design of the file format.


def binary_to_triangle(filename: str, compress: bool | None = None) -> Triangle:
    """Read a Bermuda triangle from Bermuda's binary file format.

    Args:
        filename: The filename to read the triangle from.
        compress: Whether the source file is compressed. If not provided explicitly, compression
            will be inferred from the file extension.
    """
    filepath = Path(filename).expanduser()
    extension = filepath.suffix
    # Try to deduce compression status if not provided explicitly
    if not compress:
        if extension == ".trib":
            compress = False
        elif extension == ".tribc":
            compress = True
        else:
            raise ValueError(
                f"Can't infer compression status on a triangle saved with {extension} extension"
            )
    # Check if compression status and extension are compatible if provided explicitly
    else:
        if compress and extension != ".tribc":
            warnings.warn(
                "Compressed Bermuda binaries should be saved with the `tribc` extension"
            )
        elif not compress and extension != ".trib":
            warnings.warn(
                "Uncompressed Bermuda binaries should be saved with the `trib` extension"
            )

    # Read the triangle from the appropriate kind of file
    if filename.startswith("s3:"):
        # Stream directly from S3
        raw = _open_s3_stream(filename)
        try:
            if compress:
                # Decompress as a stream; buffer again so .peek() is reliable downstream
                with gzip.GzipFile(fileobj=raw, mode="rb") as gz:
                    stream = io.BufferedReader(gz, buffer_size=64 * 1024)
                    return _read_triangle(stream)
            else:
                return _read_triangle(raw)
        finally:
            # Make sure to close the StreamingBody/BufferedReader
            raw.close()
    else:
        return _read_binary(filepath, compress)


class _BodyRawIO(io.RawIOBase):
    """Adapter so BufferedReader can wrap S3 StreamingBody."""

    def __init__(self, body):
        self.body = body

    def readable(self) -> bool:
        return True

    def readinto(self, b) -> int:
        # Read up to len(b) bytes from the StreamingBody
        data = self.body.read(len(b))
        if not data:
            return 0
        n = len(data)
        b[:n] = data
        return n

    def read(self, n: int = -1) -> bytes:
        # Optional, BufferedReader uses readinto() primarily
        return self.body.read(None if n is None or n < 0 else n)

    def close(self):
        try:
            self.body.close()
        finally:
            super().close()


def _parse_s3_uri(uri: str) -> tuple[str, str]:
    # supports s3://bucket/key and s3:bucket/key
    u = uri.replace("s3://", "s3:", 1)
    _, rest = u.split("s3:", 1)
    bucket, key = rest.lstrip("/").split("/", 1)
    return bucket, key


_S3 = None


def _open_s3_stream(s3_uri: str) -> io.BufferedReader:
    global _S3
    if _S3 is None:
        _S3 = boto3.client(
            "s3",
            config=Config(
                max_pool_connections=64,
                retries={"max_attempts": 10, "mode": "standard"},
                tcp_keepalive=True,
            ),
        )
    bucket, key = _parse_s3_uri(s3_uri)
    resp = _S3.get_object(Bucket=bucket, Key=key)
    body = resp["Body"]  # botocore.response.StreamingBody
    # Wrap in BufferedReader so .peek() works and small reads are efficient
    return io.BufferedReader(_BodyRawIO(body), buffer_size=64 * 1024)


def _read_binary(filepath, compress):
    if compress:
        with gzip.open(filepath, "rb", compresslevel=5) as infile:
            return _read_triangle(infile)
    else:
        with open(filepath, "rb") as infile:
            return _read_triangle(infile)


def _read_triangle(stream) -> Triangle:
    # Make sure the file has the right magic and version number
    if stream.read(4) != MAGIC:
        raise ValueError("This file is not a Bermuda triangle blob")
    if stream.read(1) != VERSION:
        raise ValueError(
            "The version of this file is not supported by this version of Bermuda"
        )

    # Read the string pool
    string_pool = _read_string_pool(stream)

    # Read the triangle contents
    current_metadata = None
    cells = []
    while True:
        marker = stream.read(1)
        # Only need to read the metadata if it's explicitly declared
        if marker == METADATA:
            current_metadata = _read_metadata(stream, string_pool)
        elif marker == CELL or marker == CUMULATIVE_CELL or marker == INCREMENTAL_CELL:
            cells.append(_read_cell(stream, marker, current_metadata, string_pool))
        # stream.read(1) will return an empty array if we're at EOF
        else:
            break

    return Triangle(cells)


def _read_string_pool(stream) -> list[str]:
    # Read the length of the string pool
    num_elements = struct.unpack("<H", stream.read(2))[0]
    # Construct an array with all of the string pool elements
    return [_read_string(stream) for _ in range(num_elements)]


def _read_cell(
    stream, marker: bytes, metadata: Metadata, string_pool: list[str]
) -> Cell:
    # Construct the appropriate type of cell based on the Cell-type marker byte
    if marker == CUMULATIVE_CELL:
        return CumulativeCell(
            period_start=_read_date(stream),
            period_end=_read_date(stream),
            evaluation_date=_read_date(stream),
            values=_read_dict(stream, string_pool),
            metadata=metadata,
        )
    elif marker == INCREMENTAL_CELL:
        return IncrementalCell(
            period_start=_read_date(stream),
            period_end=_read_date(stream),
            evaluation_date=_read_date(stream),
            values=_read_dict(stream, string_pool),
            prev_evaluation_date=_read_date(stream),
            metadata=metadata,
        )
    else:
        return Cell(
            period_start=_read_date(stream),
            period_end=_read_date(stream),
            evaluation_date=_read_date(stream),
            values=_read_dict(stream, string_pool),
            metadata=metadata,
        )


def _read_metadata(stream, string_pool: list[str]) -> Metadata:
    # Construct a Metadata object from the serialized fields
    return Metadata(
        risk_basis=_read_string(stream),
        country=_read_string(stream),
        currency=_read_string(stream),
        reinsurance_basis=_read_string(stream),
        loss_definition=_read_string(stream),
        per_occurrence_limit=_read_float(stream),
        details=_read_dict(stream, string_pool),
        loss_details=_read_dict(stream, string_pool),
    )


def _read_string(stream) -> str | None:
    # Read the string length
    length = struct.unpack("<h", stream.read(2))[0]
    # Special handling for None values
    if length == -1:
        return None
    return stream.read(length).decode("utf-8")


def _read_date(stream) -> datetime.date:
    year, month, day = struct.unpack("<hBB", stream.read(4))
    return datetime.date(year, month, day)


def _read_float(stream) -> float | None:
    num = struct.unpack("<d", stream.read(8))[0]
    # Special handling for None values
    if math.isnan(num):
        return None
    return num


def _read_array(stream, dtype: np.dtype) -> np.ndarray:
    # Read the number of dimensions
    num_dims = struct.unpack("<B", stream.read(1))[0]
    # Read the array shape
    shape = tuple([struct.unpack("<L", stream.read(4))[0] for _ in range(num_dims)])

    # Figure out the total number of items in the array
    num_items = 1
    for dim in shape:
        num_items *= dim

    # Read the array values
    # Not using np.fromfile() because NumPy uses seek() internally, which doesn't play
    # nicely with compression
    array_bytes = stream.read(num_items * 8)
    return np.frombuffer(array_bytes, dtype).reshape(shape)


def _read_dict(stream, string_pool: list[str]) -> dict[str, MetadataValue]:
    result = {}
    # Keep reading key-value pairs
    while stream.peek(1)[:1] != DICT_END:
        # Grab the index of the key in the string pool
        key_ndx = struct.unpack("<H", stream.read(2))[0]
        key = string_pool[key_ndx]
        # Read the value
        result[key] = _read_generic_value(stream)
    # Consume the end-of-dict byte that we must have just peeked at
    stream.read(1)
    return result


def _read_generic_value(stream) -> MetadataValue:
    # Read the value type byte
    valtype = stream.read(1)
    # Read the contextually appropriate value type, based on the marker byte
    if valtype == STRING:
        return _read_string(stream)
    elif valtype == BOOL:
        return struct.unpack("?", stream.read(1))[0]
    elif valtype == INT:
        return struct.unpack("<q", stream.read(8))[0]
    elif valtype == FLOAT:
        return struct.unpack("<d", stream.read(8))[0]
    elif valtype == INT_ARRAY:
        return _read_array(stream, np.dtype("int64"))
    elif valtype == FLOAT_ARRAY:
        return _read_array(stream, np.dtype("float64"))
    elif valtype == DATE:
        return _read_date(stream)
    elif valtype == NONE:
        return None



import datetime
import gzip
import math
import struct
import warnings
from pathlib import Path
from tempfile import NamedTemporaryFile

import awswrangler as wr
import numpy as np

from ..base import Cell, CumulativeCell, IncrementalCell, Metadata, MetadataValue
from ..triangle import Triangle
from .binary import *

__all__ = [
    "triangle_to_binary",
]


# Here's a quick crash course on binary file formats in general, and Bermuda binary in particular.
#
# All of this commentary is going to be written from the perspective of writing a binary file,
# but it applies just as well to reading (except in reverse, of course). The writer and reader
# are near-exact mirror images of each other -- for every _write_* function in this file, there's
# a corresponding inverse _read_* function in binary_input.py. Reading both files at the same
# time in adjacent windows is not a bad idea (and in fact is how I wrote them).
#
# Bermuda binary is a byte-aligned format, meaning that every token that we're going to be writing
# is a multiple of 8 bits. We could get a little bit of space savings by only using exactly the
# number of bits we need at each point, but only at the cost of quite a lot of additional
# complexity. Conceptually, we can think of the files an array of bytes (which is not the same
# thing as a string!). We just need a way to convert triangles and their constituent types to
# bytearrays and vice versa.
#
# We rely on the struct package to convert primitives (like integers) to or from bytearrays.
# Unlike vanilla Python, we don't have (easy) access to infinite-precision integers. For the
# sake of efficiency, we're using fixed-width types for all integer values, which means that
# there are limits on what can be serialized. For example, strings must be less than 32,768
# characters long (unless they contain non-ASCII characters, where the limit is lower). In
# general, the file format size limits are set high enough that I can't conceive of any future
# non-perverse use case that would exceed them.
#
# We specify little-endian byte ordering for (almost) everything that we serialize. Unfortunately,
# that's not sufficient to ensure cross-platform compatibility, as arrays are written with
# platform-native byte ordering performance reasons. However, x86, x86-64, and ARM (i.e. M1)
# are all little-endian, so we should be safe until we start trying to run Bermuda on a
# mainframe with IBM AIX processors.
#
# Higher-order data types are straightforward extensions of primitives. We serialize strings
# by writing a two-byte field with the number of bytes in the UTF-8-encoded string, followed
# by the UTF-8-encoded string itself. We serialize dictionaries by writing a series of key-value
# pairs, followed by a special byte that indicates the end of the dictionary. We serialize
# polymorphic types (such as dictionary values, which can be arrays, strings, floats, ints,
# booleans, etc) by writing a special byte that indicates the data type, followed by the binary
# representation of the value itself. We serialize dates by writing out the year, month, and day
# as fixed-width integers. We serialize objects (such as Metadata and Cell) by writing out all
# of their fields in a specified order.
#
# The Bermuda binary layout is pretty straightforward. Writing a binary consists of the following
# steps:
#   - Write the magic number (see the Wikipedia article at File_format#Magic_number
#   - Write the version number of the file format (which is 1 for now; if we ever need to
#     change the format, this allows us to preserve backwards compatibility)
#   - Write the string pool. This is the set of all distinct strings that show up as keys in
#     cell.values, metadata.details, or metadata.loss_details. This lets us save space when
#     writing dictionaries. Instead of writing out the string "earned_premium" hundreds of times,
#     we can simply write the index in the string pool of the "earned_premium" entry each time.
#   - For each cell in the triangle, see if the cell has the same metadata as the previous cell.
#     (If the cell is the first cell, assume the previous cell metadata is None.) If the metadata
#     is different, write a metadata indicator byte, then write the contents of the new metadata.
#     Regardless, then write a byte that indicates the type of cell, followed by the contents of
#     the cell except for the metadata. If cell is not immediately preceded by metadata, we can
#     assume that the cell's metadata is the same as the previous cell.


def triangle_to_binary(
    triangle: Triangle, filename: str, compress: bool = False
) -> None:
    """Convert a Bermuda triangle to a binary file format.

    Args:
        triangle: The Bermuda triangle to save out.
        filename: The filename to save the triangle to.
        compress: Whether to compress the file.
    """
    # Make sure the filename extension follows our naming conventions.
    filepath = Path(filename).expanduser()
    extension = filepath.suffix

    if compress and extension != ".tribc":
        warnings.warn(
            "Compressed Bermuda binaries should be saved with the `tribc` extension"
        )
    elif not compress and extension != ".trib":
        warnings.warn(
            "Uncompressed Bermuda binaries should be saved with the `trib` extension"
        )

    # Write the triangle to the appropriate kind of file
    if filename.startswith("s3:"):
        with NamedTemporaryFile() as temp:
            _write_binary(triangle, temp.name, compress)
            wr.s3.upload(local_file=temp.name, path=filename, use_threads=True)
    else:
        _write_binary(triangle, filepath, compress)


def _write_binary(triangle, filepath, compress):
    if compress:
        with gzip.open(filepath, "wb", compresslevel=5) as outfile:
            _write_triangle(triangle, outfile)
    else:
        with open(filepath, "wb") as outfile:
            _write_triangle(triangle, outfile)


def _write_triangle(triangle: Triangle, stream):
    # Write the file header
    stream.write(MAGIC)
    stream.write(VERSION)

    # Write the string pool
    pool_lookup = _write_string_pool(triangle, stream)

    # Write the triangle contents
    prev_metadata = None
    for cell in triangle:
        # Only need to write the metadata if it's different from the previous cell
        if prev_metadata != cell.metadata:
            _write_metadata(cell.metadata, stream, pool_lookup)
        _write_cell(cell, stream, pool_lookup)
        prev_metadata = cell.metadata


def _write_string_pool(triangle: Triangle, stream) -> dict[str, int]:
    # Collect all of the keys used in all of the dictionaries
    all_keys = set(triangle.fields)
    for metadata in triangle.metadata:
        all_keys |= set(metadata.details.keys()) | set(metadata.loss_details.keys())
    # Explicitly sorting here so the hash of a Bermuda binary file for a fixed set of contents
    # will be invariant.
    all_sorted_keys = sorted(all_keys)

    # Write out the length of the string pool
    stream.write(struct.pack("<h", len(all_sorted_keys)))

    # Write out the constituent strings and construct a string-to-index lookup table
    pool_lookup = dict()
    for ndx, key in enumerate(all_sorted_keys):
        _write_string(key, stream)
        pool_lookup[key] = ndx

    return pool_lookup


def _write_cell(cell: Cell, stream, pool_lookup: dict[str, int]):
    # Write the byte that indicates what type of Cell it is
    if isinstance(cell, CumulativeCell):
        stream.write(CUMULATIVE_CELL)
    elif isinstance(cell, IncrementalCell):
        stream.write(INCREMENTAL_CELL)
    else:
        stream.write(CELL)

    # Write the cell contents
    _write_date(cell.period_start, stream)
    _write_date(cell.period_end, stream)
    _write_date(cell.evaluation_date, stream)
    _write_dict(cell.values, stream, pool_lookup)
    if isinstance(cell, IncrementalCell):
        _write_date(cell.prev_evaluation_date, stream)


def _write_metadata(metadata: Metadata, stream, pool_lookup: dict[str, int]):
    # Write the metadata indicator byte
    stream.write(METADATA)

    # Write all of the metadata contents
    _write_string(metadata.risk_basis, stream)
    _write_string(metadata.country, stream)
    _write_string(metadata.currency, stream)
    _write_string(metadata.reinsurance_basis, stream)
    _write_string(metadata.loss_definition, stream)
    _write_float(metadata.per_occurrence_limit, stream)
    _write_dict(metadata.details, stream, pool_lookup)
    _write_dict(metadata.loss_details, stream, pool_lookup)


def _write_string(string: str | None, stream):
    # Special handling for None values
    if string is None:
        stream.write(struct.pack("<h", -1))
    else:
        # Implicit encoding to UTF-8 here
        string_bytes = string.encode()
        # Write the number of bytes -- not guaranteed to be the same as len(string)!
        stream.write(struct.pack("<H", len(string_bytes)))
        # Write the string contents
        stream.write(string_bytes)


def _write_date(date: datetime.date, stream):
    # Signed 2-byte for year, unsigned 1-byte for month and day
    stream.write(struct.pack("<hBB", date.year, date.month, date.day))


def _write_float(num: float | None, stream):
    # Use nan to represent Nones
    if num is None:
        stream.write(struct.pack("<d", math.nan))
    else:
        stream.write(struct.pack("<d", num))


def _write_array(array: np.ndarray, stream):
    # Write a byte indicating the dtype of the array
    if array.dtype == "float64":
        stream.write(FLOAT_ARRAY)
    elif array.dtype == "int64":
        stream.write(INT_ARRAY)
    else:
        raise ValueError(f"Can't serialize a NumPy array of dtype {array.dtype}")

    # Write the number of dimensions in the array
    num_dims = len(array.shape)
    stream.write(struct.pack("<B", num_dims))
    # Write each of the dimensions in the array
    for dim in array.shape:
        stream.write(struct.pack("<L", dim))
    # Write the array values
    # Not using array.tofile() because NumPy uses seek() internally, which doesn't
    # play nicely with compression
    stream.write(array.tobytes())


def _write_dict(dct: dict[str, MetadataValue], stream, pool_lookup: dict[str, int]):
    # Write each of the key-value pairs
    for k, v in dct.items():
        # Write the index of the key in the string pool
        stream.write(struct.pack("<H", pool_lookup[k]))
        # Write the value
        _write_generic_value(v, stream)
    # Add the end-of-dict byte marker
    stream.write(DICT_END)


def _write_generic_value(val: MetadataValue, stream):
    # Figure out the data type, write a byte indicating the value type, then write the
    # value itself.
    if isinstance(val, str):
        stream.write(STRING)
        _write_string(val, stream)
    elif isinstance(val, bool):
        stream.write(BOOL)
        stream.write(struct.pack("?", val))
    elif isinstance(val, (int, np.int64)):
        stream.write(INT)
        stream.write(struct.pack("<q", val))
    elif isinstance(val, float):
        stream.write(FLOAT)
        stream.write(struct.pack("<d", val))
    elif isinstance(val, np.ndarray):
        _write_array(val, stream)
    elif isinstance(val, datetime.date):
        stream.write(DATE)
        _write_date(val, stream)
    else:
        stream.write(NONE)



from .cell import Cell, CellValue
from .incremental import CumulativeCell, IncrementalCell
from .metadata import Metadata, MetadataValue, common_metadata, metadata_diff

__all__ = [
    "Metadata",
    "MetadataValue",
    "common_metadata",
    "metadata_diff",
    "Cell",
    "CellValue",
    "IncrementalCell",
    "CumulativeCell",
]



import copy
import datetime
from collections import defaultdict

import toolz as tlz

from ..base import CumulativeCell, IncrementalCell
from ..date_utils import add_months, id_to_month, month_to_id, resolution_delta
from ..errors import TriangleError
from ..triangle import Triangle

__all__ = [
    "to_incremental",
    "to_cumulative",
]


def to_incremental(
    triangle: Triangle[CumulativeCell],
) -> Triangle[IncrementalCell]:
    """Convert a Triangle from cumulative to incremental basis."""

    if triangle.is_incremental:
        return triangle

    grouped_cells = tlz.groupby(lambda ob: (ob.period, ob.metadata), triangle.cells)
    ordered_cells = tlz.valmap(sorted, grouped_cells)

    incremental_cells = []
    for ((period_start, period_end), metadata), cells in ordered_cells.items():
        # Add the first cell
        incremental_cells.append(
            IncrementalCell(
                period_start=period_start,
                period_end=period_end,
                prev_evaluation_date=resolution_delta(period_start, (-1, "days")),
                evaluation_date=cells[0].evaluation_date,
                metadata=metadata,
                values=copy.deepcopy(cells[0].values),
            )
        )

        # Add subsequent cells
        for prev_cell, next_cell in zip(cells[:-1], cells[1:]):
            incremental_cells.append(
                IncrementalCell(
                    period_start=period_start,
                    period_end=period_end,
                    prev_evaluation_date=prev_cell.evaluation_date,
                    evaluation_date=next_cell.evaluation_date,
                    metadata=metadata,
                    values=_values_diff(prev_cell.values, next_cell.values),
                )
            )

    return Triangle(incremental_cells)


def to_cumulative(
    triangle: Triangle[IncrementalCell],
) -> Triangle[CumulativeCell]:
    """Convert a Triangle from incremental to cumulative basis."""

    if not triangle.is_incremental:
        return triangle

    grouped_cells = tlz.groupby(lambda ob: (ob.period, ob.metadata), triangle.cells)
    ordered_cells = tlz.valmap(sorted, grouped_cells)

    cumulative_cells = []
    for ((period_start, period_end), metadata), cells in ordered_cells.items():
        if (
            resolution_delta(cells[0].prev_evaluation_date, (1, "days"))
            != cells[0].period_start
        ):
            raise TriangleError(
                "Cannot convert incomplete incremental triangle to cumulative"
            )

        current_values = copy.deepcopy(cells[0].values)
        current_evaluation_date = cells[0].evaluation_date
        cumulative_cells.append(
            CumulativeCell(
                period_start=period_start,
                period_end=period_end,
                evaluation_date=cells[0].evaluation_date,
                metadata=metadata,
                values=copy.deepcopy(current_values),
            )
        )

        for cell in cells[1:]:
            if cell.prev_evaluation_date != current_evaluation_date:
                raise TriangleError(
                    "Cannot convert incomplete incremental triangle to cumulative"
                )
            current_evaluation_date = cell.evaluation_date
            current_values = _values_add(current_values, cell.values)
            cumulative_cells.append(
                CumulativeCell(
                    period_start=period_start,
                    period_end=period_end,
                    evaluation_date=cell.evaluation_date,
                    metadata=metadata,
                    values=current_values,
                )
            )

    return Triangle(cumulative_cells)


def accident_quarter_to_policy_year(
    accident_quarter_tri: Triangle,
    policy_length_months: int = 12,
    policy_year_origin: datetime.date = datetime.date(2020, 1, 1),
    continuous_issuance: bool = True,
) -> Triangle:
    """Converts accident quarter triangles to policy year triangles.

    Note this function will fail if the evaluation dates are not consistent for the whole policy
    period.

    Args:
        accident_quarter_tri: The triangle to convert.
        policy_length_months: The length of the policy in months.
        policy_year_origin: The date of the start of the policy year.
        continuous_issuance: Whether the policy is issued continuously or not. If not, all premium
        is written in the first month of the policy year.
    """
    results = Triangle([])
    for _, tri_slice in accident_quarter_tri.slices.items():
        results += _accident_quarter_to_policy_year_slice(
            tri_slice,
            policy_length_months=policy_length_months,
            policy_year_origin=policy_year_origin,
            continuous_issuance=continuous_issuance,
        )
    return results


def _accident_quarter_to_policy_year_slice(
    accident_quarter_tri_slice: Triangle,
    policy_length_months: int = 12,
    policy_year_origin: datetime.date = datetime.date(2020, 1, 1),
    continuous_issuance: bool = True,
) -> Triangle:
    """Converts accident quarter triangles to policy year triangles.

    Note this function will fail if the evaluation dates are not consistent for the whole policy
    period.

    Args:
        accident_quarter_tri: The triangle to convert.
        policy_length_months: The length of the policy in months.
        policy_year_origin: The date of the start of the policy year.
        continuous_issuance: Whether the policy is issued continuously or not. If not, all premium
        is written in the first month of the policy year.
    """

    if len(accident_quarter_tri_slice.metadata) > 1:
        raise ValueError("The accident quarter triangle should have only one slice.")
    if len(accident_quarter_tri_slice.right_edge.evaluation_dates) > 1:
        raise ValueError(
            "The accident quarter triangle has more than one evaluation date at the right edge. \n"
            "This will could lead to undercounting policy-basis premium and loss."
        )

    policy_years = policy_years_covered(accident_quarter_tri_slice, policy_year_origin)
    policy_year_ep_shares = {}
    for policy_year in policy_years:
        policy_year_ep_shares[policy_year] = monthly_ep_to_quarterly_ep(
            _policy_earned_premium_share_by_month(
                risk_start_date=policy_year[0],
                risk_end_date=policy_year[1],
                policy_length_months=policy_length_months,
                continuous_issuance=continuous_issuance,
            ),
            accident_quarter_tri_slice,
        )
    accident_quarter_py_shares = defaultdict(dict)
    for py, aq_share in policy_year_ep_shares.items():
        for aq, share in aq_share.items():
            accident_quarter_py_shares[aq].update({py: share})

    # normalize the shares for each accident quarter
    for _, py_shares in accident_quarter_py_shares.items():
        total_weight = sum(py_shares.values())
        for py, share in py_shares.items():
            py_shares[py] = share / total_weight

    cells = []
    for policy_period in policy_years:
        for evaluation_date in accident_quarter_tri_slice.evaluation_dates:
            aq_cells = accident_quarter_tri_slice[:, evaluation_date, :]
            vals_dict = defaultdict(float)
            for cell in aq_cells:
                py_share = accident_quarter_py_shares[cell.period]
                if policy_period in py_share:
                    for field, val in cell.values.items():
                        vals_dict[field] += val * py_share[policy_period]
            if vals_dict:
                cells.append(
                    CumulativeCell(
                        period_start=policy_period[0],
                        period_end=policy_period[1],
                        evaluation_date=evaluation_date,
                        values=vals_dict,
                        metadata=cell.metadata,
                    )
                )

    return Triangle(cells).derive_metadata(risk_basis="Policy")


def _policy_earned_premium_share_by_month(
    risk_start_date: datetime.date,
    risk_end_date: datetime.date,
    policy_length_months: int,
    continuous_issuance: bool,
    monthly_earning_pattern: dict[datetime.date, float] | None = None,
) -> dict[datetime.date, float]:
    written_start_month_id = month_to_id(risk_start_date)
    if continuous_issuance:
        written_end_month_id = month_to_id(risk_end_date)
    else:
        written_end_month_id = written_start_month_id

    earned_premium_by_month = {
        month_id: 0
        for month_id in range(
            written_start_month_id, written_end_month_id + policy_length_months + 1
        )
    }
    if monthly_earning_pattern:
        for month_id in earned_premium_by_month:
            earned_premium_by_month[month_id] = {
                month_to_id(date): value
                for date, value in monthly_earning_pattern.items()
            }.get(month_id, 0)
    else:
        written_months = written_end_month_id - written_start_month_id + 1
        monthly_volume = 1 / written_months / policy_length_months
        for written_month_id in range(written_start_month_id, written_end_month_id + 1):
            earned_premium_by_month[written_month_id] += monthly_volume / 2
            for earn_offset in range(1, policy_length_months):
                earned_premium_by_month[written_month_id + earn_offset] += (
                    monthly_volume
                )
            earned_premium_by_month[written_month_id + policy_length_months] += (
                monthly_volume / 2
            )

    return {id_to_month(id_): value for id_, value in earned_premium_by_month.items()}


def policy_years_covered(
    accident_basis_tri: Triangle, policy_year_origin: datetime.date
) -> list[tuple[datetime.date, datetime.date]]:
    """Returns the policy years covered by the accident-basis triangle."""
    first_accident_period_start = accident_basis_tri.periods[0][0]
    last_accident_period_end = accident_basis_tri.periods[-1][1]
    py_start = datetime.date(
        first_accident_period_start.year,
        policy_year_origin.month,
        policy_year_origin.day,
    )
    if py_start > first_accident_period_start:
        py_start = datetime.date(
            first_accident_period_start.year - 1,
            policy_year_origin.month,
            policy_year_origin.day,
        )
    period_starts = [py_start]
    while py_start < last_accident_period_end:
        py_start = add_months(py_start, 12)
        period_starts.append(py_start)
    return [
        (start, add_months(start, 12) + datetime.timedelta(days=-1))
        for start in period_starts
    ]


def monthly_ep_to_quarterly_ep(
    policy_month_ep_share: dict[datetime.date, float], accident_tri: Triangle
) -> dict[tuple, float]:
    """Aggregates the monthly EP share dictionary to match the quarterly periods."""
    accident_quarter_ep_share = defaultdict(float)
    for month, share in policy_month_ep_share.items():
        for quarter in accident_tri.periods:
            if month >= quarter[0] and month <= quarter[1]:
                accident_quarter_ep_share[quarter] += share
    return accident_quarter_ep_share


def _values_diff(prev_values, next_values):
    prev_keys = set(prev_values.keys())
    next_keys = set(next_values.keys())
    if prev_keys.symmetric_difference(next_keys):
        raise TriangleError("Consecutive cells have different value keys")

    return {
        k: (
            next_values[k] if k == "earned_premium" else next_values[k] - prev_values[k]
        )
        for k in prev_keys
    }


def _values_add(curr_values, next_values):
    curr_keys = set(curr_values.keys())
    next_keys = set(next_values.keys())
    if curr_keys.symmetric_difference(next_keys):
        raise TriangleError("Consecutive cells have different value keys")

    return {
        k: (
            next_values[k] if k == "earned_premium" else curr_values[k] + next_values[k]
        )
        for k in curr_keys
    }



import datetime
from collections import namedtuple
from typing import Any, Tuple, Union

import numpy as np

from ..date_utils import calculate_dev_lag
from .cell import Cell, CellValue, format_value
from .metadata import Metadata

__all__ = [
    "CumulativeCell",
    "IncrementalCell",
]


_constr_coordinates = namedtuple(
    "coordinates",
    ["period_start", "period_end", "evaluation_date", "prev_evaluation_date"],
)


class CumulativeCell(Cell):
    """Subclass for cumulative cells.

    The base Cell class is already cumulative, so this subclass mostly exists for the sake of the
    type system. Cell implies cumulative or incremental; CumulativeCell implies that it doesn't
    work with incremental cells.
    """

    pass


class IncrementalCell(Cell):
    def __init__(
        self,
        period_start: datetime.date,
        period_end: datetime.date,
        prev_evaluation_date: datetime.date,
        evaluation_date: datetime.date,
        values: dict[str, CellValue],
        metadata: Metadata = None,
        _skip_validation: bool = False,
    ):
        super().__init__(
            period_start=period_start,
            period_end=period_end,
            evaluation_date=evaluation_date,
            values=values,
            metadata=metadata,
            _skip_validation=_skip_validation,
        )
        if evaluation_date <= prev_evaluation_date and not _skip_validation:
            raise ValueError(
                f"`evaluation_date` ({evaluation_date}) must be > `prev_evaluation_date` ({prev_evaluation_date})."
            )
        self._prev_evaluation_date = prev_evaluation_date

    prev_evaluation_date = property(lambda self: self._prev_evaluation_date)

    def __eq__(self, other: Cell) -> bool:
        return (
            super().__eq__(other)
            and self._prev_evaluation_date == other._prev_evaluation_date
        )

    def __lt__(self, other: Cell) -> bool:
        """Required so that lists of cells can be sortable.
        Comparison is based on (period_start, period_end, evaluation_date, prev_evaluation_date) tuple.
        """
        return (
            self._metadata,
            self._period_start,
            self._period_end,
            self._evaluation_date,
            self._prev_evaluation_date,
        ) < (
            other._metadata,
            other._period_start,
            other._period_end,
            other._evaluation_date,
            other._prev_evaluation_date,
        )

    @property
    def coordinates(
        self,
    ) -> Tuple[datetime.date, datetime.date, datetime.date, datetime.date]:
        """Location of the cell: period_start, period_end, evaluation_date, prev_evaluation_date."""
        return _constr_coordinates(
            self._period_start,
            self._period_end,
            self._evaluation_date,
            self._prev_evaluation_date,
        )

    def eval_lag(self, unit: str = "months") -> Union[float, int, datetime.timedelta]:
        """Compute the evaluation lag of the cell in the specified units.

        Args:
            unit (str): One of 'month', 'day', 'timedelta'.  'month' returns eval_lag as
                a float number of months, day returns dev_lag as an integer number
                of days, and 'timedelta' returns a datetime.timedelta object.
        """
        return calculate_dev_lag(
            self._prev_evaluation_date, self._evaluation_date, unit
        )

    def to_record(self, dev_lag_unit="month") -> dict[str, Any]:
        """Return dict representation of Cell."""
        return {
            "period_start": self._period_start,
            "period_end": self._period_end,
            "evaluation_date": self._evaluation_date,
            "prev_evaluation_date": self._prev_evaluation_date,
            "dev_lag": self.dev_lag(dev_lag_unit),
            **self._metadata.as_flat_dict(),
            **self._values,
        }

    def _repr_html_(self) -> str:
        metadata_rows = []
        for name, value in self.metadata.as_flat_dict().items():
            if value is not None:
                metadata_rows.append(f"<tr><th> {name} </th> <td> {value} </td></tr>")

        value_rows = []
        for name, value in self.values.items():
            if isinstance(value, np.ndarray):
                # noinspection PyTypeChecker
                value_summary = f"~{format_value(np.mean(value))} (Size {value.size})"
            else:
                value_summary = format_value(value)
            value_rows.append(f"<tr><th> {name} </th> <td> {value_summary} </td></tr>")

        table = [
            "<table>",
            "  <thead>",
            f"    <tr><th> <h4><b>{self.__class__.__name__}</b></h4> </th></tr>",
            "  </thead>",
            "  <tbody>",
            f"    <tr><th> Period Start: </th><td> {self.period_start} </td></tr>",
            f"    <tr><th> Period End: </th><td> {self.period_end} </td></tr>",
            f"    <tr><th> Evaluation Date: </th><td> {self.evaluation_date} </td></tr>",
            f"    <tr><th> Previous Evaluation Date: </th><td> {self.prev_evaluation_date} </td></tr>",
            f"    <tr><th> Development Lag: </th><td> {self.dev_lag()} months </td></tr>",
            '    <tr><th colspan=2 style="text-align:center"> Values </th></tr>',
            *value_rows,
            '    <tr><th colspan=2 style="text-align:center"> Metadata </th></tr>',
            *metadata_rows,
            "  </tbody></table>",
        ]
        return "\n".join(table)



from typing import Tuple

import toolz as tlz

from ..base import Cell, Metadata
from ..triangle import Triangle

__all__ = [
    "join",
]


def join(
    tri1: Triangle,
    tri2: Triangle,
    join_type: str = "full",
    on: list[str] | None = None,
) -> list[Tuple[Cell | None, Cell | None]]:
    """Create pairs of cells that share the same coordinates and metadata from two triangles.

    Join is a lower-level operation that returns a list of pairs of cells. Many common use cases
    are better served by its higher-order cousin, `merge`.

    Args:
        tri1: The left triangle in the join.
        tri2: The right triangle in the join.
        join_type: The type of relational join to perform. Available join types are:
            * inner: Return only cell coordinates that are present in both triangles.
            * left: Return all cell coordinates in the left triangle, along with any
                available matches on the right triangle.
            * right: Return all cell coordinates in the right triangle, along with any
                available matches on the left triangle.
            * full: Return all cell coordinates in both triangles, regardless of
                whether there is a match on the other side.
            * left_anti: Return only those cells in the left triangle that do not have
                matching coordinates in the right triangle.
            * right_anti: Return only those cells in the right triangle that do not
                have matching coordinates in the left triangle.
        on: which metadata attributes and detail keys to join on when merging cell values.
            Default behavior is to consider all metadata attributes when merging cells.
    Returns:
        A list of 2-tuples of `Cell`s. Each tuple represents a pair of cells with the
        same coordinates and metadata. The first and second elements of the tuple are
        the cell from the left and right triangles, respectively. Depending on the
        type of join, either of the cells in the tuple may be `None`, but by construction,
        every tuple will have at least one non-`None` cell.
    """
    # Make sure that both triangles have the same cell type
    if min(len(tri1), len(tri2)) > 0 and type(tri1.cells[0]) != type(tri2.cells[0]):  # noqa: E721
        raise ValueError("Triangles must have the same cell type")

    # If attributes are passed to `on` we will ignore any other metadata by removing all other
    # metadata attributes
    if on:
        tri1 = _select_metadata(tri1, on)
        tri2 = _select_metadata(tri2, on)

    # Create an index of all cells in each triangle
    if tri1.is_incremental:
        tri1_cells = {
            (
                cell.metadata,
                cell.period_start,
                cell.period_end,
                cell.evaluation_date,
                cell.prev_evaluation_date,
            ): cell
            for cell in tri1
        }
        tri2_cells = {
            (
                cell.metadata,
                cell.period_start,
                cell.period_end,
                cell.evaluation_date,
                cell.prev_evaluation_date,
            ): cell
            for cell in tri2
        }
    else:
        tri1_cells = {
            (
                cell.metadata,
                cell.period_start,
                cell.period_end,
                cell.evaluation_date,
            ): cell
            for cell in tri1
        }
        tri2_cells = {
            (
                cell.metadata,
                cell.period_start,
                cell.period_end,
                cell.evaluation_date,
            ): cell
            for cell in tri2
        }

    # Merge them into a single list of 2-tuples
    all_coordinates = set(tri1_cells.keys()) | set(tri2_cells.keys())
    cell_pairs = [
        (tri1_cells.get(coord), tri2_cells.get(coord)) for coord in all_coordinates
    ]

    if join_type == "full":
        return cell_pairs
    elif join_type == "left":
        return [(cell1, cell2) for cell1, cell2 in cell_pairs if cell1 is not None]
    elif join_type == "right":
        return [(cell1, cell2) for cell1, cell2 in cell_pairs if cell2 is not None]
    elif join_type == "inner":
        return [
            (cell1, cell2)
            for cell1, cell2 in cell_pairs
            if cell1 is not None and cell2 is not None
        ]
    elif join_type == "left_anti":
        return [(cell1, cell2) for cell1, cell2 in cell_pairs if cell2 is None]
    elif join_type == "right_anti":
        return [(cell1, cell2) for cell1, cell2 in cell_pairs if cell1 is None]
    else:
        raise ValueError(f"Unrecognized join type `{join_type}`")


def _select_metadata(triangle, attributes_list):
    def metadata_mod(cell):
        new_details = tlz.keyfilter(
            lambda key: key in attributes_list, cell.metadata.details
        )
        new_loss_details = tlz.keyfilter(
            lambda key: key in attributes_list, cell.metadata.loss_details
        )
        return Metadata(
            risk_basis=(
                cell.metadata.risk_basis if "risk_basis" in attributes_list else None
            ),
            reinsurance_basis=(
                cell.metadata.reinsurance_basis
                if "reinsurance_basis" in attributes_list
                else None
            ),
            country=cell.metadata.country if "country" in attributes_list else None,
            currency=cell.metadata.currency if "currency" in attributes_list else None,
            loss_definition=(
                cell.metadata.loss_definition
                if "loss_definition" in attributes_list
                else None
            ),
            per_occurrence_limit=(
                cell.metadata.per_occurrence_limit
                if "per_occurrence_limit" in attributes_list
                else None
            ),
            details=new_details,
            loss_details=new_loss_details,
        )

    metadata_patched_cells = [
        cell.replace(metadata=metadata_mod(cell)) for cell in triangle
    ]
    return Triangle(metadata_patched_cells)



from __future__ import annotations

import calendar
import datetime
import math
from collections import Counter, defaultdict

import numpy as np

__all__ = [
    "DevLag",
    "calculate_dev_lag",
    "add_months",
    "dev_lag_months",
    "standardize_resolution",
    "resolution_delta",
    "month_to_id",
    "id_to_month",
    "period_resolution",
    "eval_date_resolution",
    "is_triangle_monthly",
    "drop_off_diagonals",
]


DevLag = int | float | datetime.timedelta


def calculate_dev_lag(
    period_end: datetime.date, evaluation_date: datetime.date, unit: str = "months"
) -> DevLag:
    """Compute the development lag of the cell in the specified units."""
    _unit = unit.lower()

    # Short-circuit handling of maximum evaluation dates
    if evaluation_date == datetime.date.max:
        if _unit == "timedelta":
            return datetime.timedelta.max
        else:
            return np.inf

    # Base case
    if "month" in _unit:
        return dev_lag_months(period_end, evaluation_date)
    elif "day" in _unit:
        return (evaluation_date - period_end).days
    elif _unit == "timedelta":
        return evaluation_date - period_end
    else:
        raise ValueError(
            f"Unrecognized unit '{unit}', must be 'month', 'day' or 'timedelta'."
        )


def add_months(dt: datetime.date, delta: float) -> datetime.date:
    """Add a floating-point number of months to a date and return a date."""
    # Short-circuit handling of infinite deltas:
    if delta == np.inf:
        return datetime.date.max

    # Compute result as an offset from a fixed origin date
    init_lag = dev_lag_months(datetime.date(1969, 12, 31), dt)
    final_lag = init_lag + delta

    # Split the result into an integer and fractional part, with the convention that the
    # fractional part is in the domain (0.0, 1.0].
    final_lag_months = int(final_lag) - 1 if final_lag % 1 == 0.0 else int(final_lag)
    final_lag_frac = 1.0 if final_lag % 1 == 0.0 else final_lag % 1

    # Figure out the calendar year, month, and day of the result.
    month = final_lag_months % 12 + 1
    year = 1970 + final_lag_months // 12
    _, days_in_month = calendar.monthrange(year, month)
    day = round(final_lag_frac * days_in_month)

    # Special-casing required if day-of-month is zero.
    if day == 0:
        return datetime.date(year, month, 1) - datetime.timedelta(days=1)
    else:
        return datetime.date(year, month, day)


def _month_fraction(dt: datetime.date) -> float:
    """Compute the fractional portion of the month that has elapsed."""
    _, last_day_of_month = calendar.monthrange(dt.year, dt.month)
    return dt.day / last_day_of_month


def dev_lag_months(start: datetime.date, stop: datetime.date) -> float:
    """Compute the floating-point number of months between two dates."""
    start_frac = _month_fraction(start)
    stop_frac = _month_fraction(stop)
    delta_years = stop.year - start.year
    delta_months = stop.month - start.month
    # adding 0.019 to a 31 day month is the smallest value to increment by one day
    return 12 * delta_years + delta_months - start_frac + stop_frac


def standardize_resolution(resolution):
    # Convert resolution to a standard format -- units are either "day" or "month"
    quantity, units = resolution
    clean_units = units.lower()
    if "month" in clean_units:
        clean_units = "month"
    elif "quarter" in clean_units:
        quantity *= 3
        clean_units = "month"
    elif "year" in clean_units:
        quantity *= 12
        clean_units = "month"
    elif "day" in clean_units:
        clean_units = "day"
    elif "week" in clean_units:
        quantity *= 7
        clean_units = "day"
    else:
        raise ValueError(f"Unrecognized resolution unit {units}")
    return quantity, clean_units


def resolution_delta(date, resolution, negative=False):
    # Add (or subtract) a specified resolution from a given date.
    quantity, units = resolution
    if negative:
        quantity *= -1
    if units == "month":
        return add_months(date, quantity)
    else:
        return date + datetime.timedelta(days=quantity)


def month_to_id(dt: datetime.date) -> int:
    """Convert a date to an integer representing the month containing the date."""
    return 12 * (dt.year - 1970) + dt.month - 1


def id_to_month(id: int, beginning=True) -> datetime.date:
    """Convert an integer representing a month to the first date in the month."""
    if beginning:
        return datetime.date(1970 + id // 12, id % 12 + 1, 1)
    else:
        id += 1
        return datetime.date(1970 + id // 12, id % 12 + 1, 1) - datetime.timedelta(
            days=1
        )


def period_resolution(tri: Triangle) -> int:  # noqa: F821
    """Determine the biggest unit of resolution (expressed in months) that can losslessly
    represent all experience periods in a triangle."""
    period_starts, period_ends = zip(*tri.periods)
    start_months = [month_to_id(start) for start in period_starts]
    next_start_months = [month_to_id(end) + 1 for end in period_ends]
    starts_ordered = sorted(set(start_months + next_start_months))
    start_diffs = _diff(starts_ordered)
    if not start_diffs:
        return None
    return _multi_gcd(start_diffs)


def eval_date_resolution(tri: Triangle) -> int:  # noqa: F821
    """Determine the biggest unit of resolution (expressed in months) that can losslessly
    represent all evaluation dates in a triangle."""
    eval_months = [month_to_id(eval_dt) for eval_dt in tri.evaluation_dates]
    evals_ordered = sorted(eval_months)
    eval_diffs = _diff(evals_ordered)
    if not eval_diffs:
        return None
    return _multi_gcd(eval_diffs)


def drop_off_diagonals(triangle: Triangle) -> Triangle:  # noqa: F821
    """Drop off-diagonal values from a triangle.

    Only keep the most common resolution and diagonals with the most
    popular origin at that resolution.
    """
    from_previous = defaultdict(list)
    to_next = defaultdict(list)
    diffs = []
    for prev_date, next_date in zip(
        triangle.evaluation_dates[:-1], triangle.evaluation_dates[1:]
    ):
        diff = calculate_dev_lag(prev_date, next_date)
        diffs.append(diff)
        from_previous[diff].append(next_date)
        to_next[diff].append(prev_date)
    most_common_resolution = Counter(diffs).most_common(1)[0][0]
    eval_origins = [
        month_to_id(date) % most_common_resolution for date in triangle.evaluation_dates
    ]
    most_common_origin = Counter(eval_origins).most_common(1)[0][0]
    most_useful_evals = [
        date
        for date, origin in zip(triangle.evaluation_dates, eval_origins)
        if origin == most_common_origin
    ]
    return triangle.filter(lambda cell: cell.evaluation_date in most_useful_evals)


def is_triangle_monthly(tri: Triangle) -> bool:  # noqa: F821
    """Test if all periods and evaluation dates in a triangle are in monthly increments."""
    period_starts, period_ends = zip(*tri.periods)
    all_starts_aligned = all([_is_month_start(start) for start in period_starts])
    all_ends_aligned = all([_is_month_end(end) for end in period_ends])
    all_evals_aligned = all(
        [_is_month_end(eval_dt) for eval_dt in tri.evaluation_dates]
    )
    return all_starts_aligned and all_ends_aligned and all_evals_aligned


def _is_month_start(dt: datetime.date) -> bool:
    """Test if a date is at the beginning of a month."""
    return dt.day == 1


def _is_month_end(dt: datetime.date) -> bool:
    """Test if a date is at the end of a month."""
    next_day = dt + datetime.timedelta(days=1)
    return next_day.day == 1


def _multi_gcd(xs: list[int]) -> int:
    """Compute the gcd of a list of integers."""
    unique_xs = list(set(xs))
    if len(unique_xs) == 1:
        return unique_xs[0]
    # Reduce over unique xs
    result = math.gcd(unique_xs[0], unique_xs[1])
    for x in unique_xs[2:]:
        result = math.gcd(result, x)
    return result


def _diff(xs: list[int]) -> list[int]:
    """Compute the differences between a list of integers."""
    return [after - before for before, after in zip(xs[:-1], xs[1:])]



import datetime
from typing import Any, Union
from warnings import warn

import awswrangler as wr
import numpy as np
import pandas as pd
import toolz as tlz

from ..base import CumulativeCell, IncrementalCell, Metadata
from ..triangle import Triangle

__all__ = [
    "wide_csv_to_triangle",
    "long_csv_to_triangle",
    "wide_data_frame_to_triangle",
    "long_data_frame_to_triangle",
]


INDEX_CUM_COLUMNS = ["period_start", "period_end", "evaluation_date"]
INDEX_COLUMNS = INDEX_CUM_COLUMNS + ["prev_evaluation_date"]

METADATA_COLUMNS = [
    "risk_basis",
    "country",
    "currency",
    "reinsurance_basis",
    "loss_definition",
    "per_occurrence_limit",
]

CORE_SET = set(INDEX_COLUMNS + METADATA_COLUMNS + ["scenario"])


def wide_csv_to_triangle(
    file_or_fname: str,
    field_cols: list[str] | None = None,
    detail_cols: list[str] | None = None,
    loss_detail_cols: list[str] | None = None,
    metadata: Metadata | None = None,
    collapse_fields: list[str] | None = None,
    **kwargs,
) -> Union[Triangle[CumulativeCell], Triangle[IncrementalCell]]:
    """
    Convert a wide-format CSV to a `Triangle`.

    For more details on wide-format CSVs/tables, see `wide_data_frame_to_triangle` (which
    this function is a thin wrapper for).

    Args:
        file_or_fname: Name of the CSV file or a handle to a file object.
        field_cols: Names of the columns that should be interpreted as fields.
        detail_cols: Names of the columns that should be interpreted as metadata details.
            At least one of `field_cols` or `detail_cols` must be supplied.
        loss_detail_cols: Names of the detail columns that should be interpreted as metadata
            loss details, rather than standard metadata details. By default, nothing is assumed
            to be a loss detail.
        metadata: Default metadata to apply to every cell in the `Triangle` (can be overridden
            by `detail_cols`).
        collapse_fields: A list of fields to collapse to a single value, rather than loading as
            an array with multiple values. This typically applies to `earned_premium` when multiple
            scenarios are being loaded.
        **kwargs: Extra arguments to `pd.read_csv`
    """
    # Read column names first
    if file_or_fname.startswith("s3://"):
        raw_df = wr.s3.read_csv(file_or_fname, nrows=1)
    else:
        raw_df = pd.read_csv(file_or_fname, nrows=1)
    parse_dates = (
        INDEX_COLUMNS if "prev_evaluation_date" in raw_df.columns else INDEX_CUM_COLUMNS
    )
    # Read full table with matching column names
    if file_or_fname.startswith("s3://"):
        df = wr.s3.read_csv(file_or_fname, parse_dates=parse_dates, **kwargs)
    else:
        df = pd.read_csv(file_or_fname, parse_dates=parse_dates, **kwargs)
    return wide_data_frame_to_triangle(
        df, field_cols, detail_cols, loss_detail_cols, metadata, collapse_fields
    )


def long_csv_to_triangle(
    file_or_fname: str, metadata: Metadata | None = None, **kwargs
) -> Union[Triangle[CumulativeCell], Triangle[IncrementalCell]]:
    """
    Convert a long-format CSV to a `Triangle`.

    For more details on long-format CSVs/tables, see `long_data_frame_to_triangle` (which
    this function is a thin wrapper for).

    Args:
        file_or_fname: Name of the CSV file or a handle to a file object.
        detail_cols: Names of the columns that should be interpreted as metadata details.
            At least one of `field_cols` or `detail_cols` must be supplied.
        loss_detail_cols: Names of the metadata columns that should be interpreted as metadata
            loss details, rather than standard metdata details. By default, nothing is assumed
            to be a loss detail.
        **kwargs: Extra arguments to `pd.read_csv`
    """
    # Read column names first
    if file_or_fname.startswith("s3://"):
        raw_df = wr.s3.read_csv(file_or_fname, nrows=1)
    else:
        raw_df = pd.read_csv(file_or_fname, nrows=1)
    parse_dates = (
        INDEX_COLUMNS if "prev_evaluation_date" in raw_df.columns else INDEX_CUM_COLUMNS
    )
    # Read full table with matching column names
    if file_or_fname.startswith("s3://"):
        df = wr.s3.read_csv(file_or_fname, parse_dates=parse_dates, **kwargs)
    else:
        df = pd.read_csv(file_or_fname, parse_dates=parse_dates, **kwargs)
    return long_data_frame_to_triangle(df, metadata=metadata)


def wide_data_frame_to_triangle(
    df: pd.DataFrame,
    field_cols: list[str] | None = None,
    detail_cols: list[str] | None = None,
    loss_detail_cols: list[str] | None = None,
    metadata: Metadata | None = None,
    collapse_fields: list[str] | None = None,
) -> Union[Triangle[CumulativeCell], Triangle[IncrementalCell]]:
    """
    Convert a wide-format `pd.DataFrame` to a `Triangle`.

    A wide-format `DataFrame` requires the following columns to be present: `period_start`,
    `period_end`, and `evaluation_date`. When the column 'prev_evaluation_date' is present,
    a triangle with incremental cells is generated. Otherwise, all triangle cells are cumulative.

    Each row of the `DataFrame` represents a single `Cell` of the `Triangle`.
    Any columns besides the three required columns are interpreted as either fields
    (i.e., `paid_loss`, `reported_loss`) or metadata (i.e., `per_occurrence_limit`,
    `currency`, `program_name`).

    This function doesn't make any assumptions about which columns are fields and which are
    metadata based on their names. Instead, at least one of `detail_cols` or `field_cols` must be
    supplied to the function. If only one of those arguments is supplied, all other columns are
    assumed to belong to the other argument.

    The `metadata` argument can be used to supply common metadata that may not be directly
    present in the `DataFrame`. For example, every `Cell` in the `Triangle` could be
    accident-basis, gross of reinsurance, including DCC. This could be supplied by a
    single `Metadata` object in the function call instead of adding several constant columns
    to the `DataFrame`.

    Args:
        df: Pandas `DataFrame` to convert to a `Triangle`.
        field_cols: Names of the columns that should be interpreted as fields.
        detail_cols: Names of the columns that should be interpreted as metadata details.
            At least one of `field_cols` or `detail_cols` must be supplied.
        loss_detail_cols: Names of the detail columns that should be interpreted as metadata
            loss details, rather than standard metadata details. By default, nothing is assumed
            to be a loss detail.
        metadata: Default metadata to apply to every cell in the `Triangle` (can be overridden
            by `detail_cols`).
        collapse_fields: A list of fields to collapse to a single value, rather than loading as
            an array with multiple values. This typically applies to `earned_premium` when multiple
            scenarios are being loaded.
    """
    _check_index_columns(df)
    if field_cols is None and detail_cols is None:
        raise Exception("At least one of field_cols or detail_cols must be supplied")
    if field_cols is not None and detail_cols is not None:
        if len(set(field_cols) & set(detail_cols)):
            raise Exception("field_cols and detail_cols must be disjoint")
    if field_cols is None:
        field_cols = list(set(df.columns.values) - CORE_SET - set(detail_cols))
    if detail_cols is None:
        detail_cols = list(set(df.columns.values) - CORE_SET - set(field_cols))
    if loss_detail_cols is None:
        loss_detail_cols = []
    if not all([col in detail_cols for col in loss_detail_cols]):
        raise Exception("loss_detail_cols must be a subset of detail_cols")
    if collapse_fields is None:
        collapse_fields = []

    # noinspection DuplicatedCode
    if metadata is None:
        metadata = Metadata()
    cell_metadata = _create_metadata(df, metadata, detail_cols, loss_detail_cols)
    period_starts = list(df.period_start.dt.date)
    period_ends = list(df.period_end.dt.date)
    evaluation_dates = list(df.evaluation_date.dt.date)
    if "prev_evaluation_date" in df.columns:
        prev_evaluation_dates = list(df.prev_evaluation_date.dt.date)

    raw_values = {colname: _clean_list_column(df[colname]) for colname in field_cols}
    values = [
        {k: np.array(v[i]) for k, v in raw_values.items() if v[i] is not None}
        for i in range(df.shape[0])
    ]

    if "prev_evaluation_date" in df.columns:
        cells = [
            IncrementalCell(
                period_start=period_start,
                period_end=period_end,
                evaluation_date=evaluation_date,
                prev_evaluation_date=prev_evaluation_date,
                metadata=metadata,
                values=value,
            )
            for (
                metadata,
                period_start,
                period_end,
                evaluation_date,
                prev_evaluation_date,
                value,
            ) in zip(
                cell_metadata,
                period_starts,
                period_ends,
                evaluation_dates,
                prev_evaluation_dates,
                values,
            )
        ]
    else:
        if "risk_basis" not in df.columns:
            df = df.assign(risk_basis=[cell.risk_basis for cell in cell_metadata])
        if "per_occurrence_limit" not in df.columns:
            df = df.assign(
                per_occurrence_limit=[
                    cell.per_occurrence_limit for cell in cell_metadata
                ]
            )
        cells = []
        for idx, group_df in df.groupby(
            [
                "period_start",
                "period_end",
                "evaluation_date",
                "risk_basis",
                "per_occurrence_limit",
            ]
            + detail_cols
            + loss_detail_cols,
            dropna=False,
        ):
            period_start, period_end, evaluation_date, *_ = idx
            current_metadata = _create_metadata(
                group_df, metadata, detail_cols, loss_detail_cols
            )[0]
            if len(group_df) > 1:
                try:
                    group_df = group_df.sort_values(["scenario"])
                except KeyError:
                    raise Exception(
                        "Multiple values found for this Cell. If multiple values are present for a single field they must be identified by a `scenario` column."
                    )
            values = {
                colname: _clean_list_column(group_df[colname]) for colname in field_cols
            }

            # Check for scalar values, typically earned_premium
            for field in collapse_fields:
                if all(val == values[field][0] for val in values[field]):
                    values[field] = values[field][0]
                else:
                    raise warn(
                        f"Field {field} has multiple distinct values, will not be collapsed"
                    )

            # coerce values types
            for field, vals in values.items():
                if isinstance(vals, list) and len(vals) > 1:
                    values[field] = np.array(vals)
                if isinstance(vals, list) and len(vals) == 1:
                    values[field] = vals[0]
            values = tlz.valfilter(lambda val: val is not None, values)

            cells.append(
                CumulativeCell(
                    period_start=period_start.date(),
                    period_end=period_end.date(),
                    evaluation_date=evaluation_date.date(),
                    metadata=current_metadata,
                    values=values,
                )
            )

    return Triangle(cells)


def long_data_frame_to_triangle(
    df: pd.DataFrame,
    loss_detail_cols: list[str] | None = None,
    metadata: Metadata | None = None,
) -> Union[Triangle[CumulativeCell], Triangle[IncrementalCell]]:
    """
    Convert a long-format `pd.DataFrame` to a `Triangle`.

    A long-format `DataFrame` requires the following columns to be present: `period_start`,
    `period_end`, `evaluation_date`, `field` and `value`. When the column 'prev_evaluation_date'
    is present, a triangle with incremental cells is generated. Otherwise, all triangle cells
    are cumulative.

    Each row of the `DataFrame` represents a field of a single `Cell` of the `Triangle`.
    Any columns besides the three required columns are interpreted as metadata
    (i.e., `per_occurrence_limit`, `currency`, `program_name`).

    The `metadata` argument can be used to supply common metadata that may not be directly
    present in the `DataFrame`. For example, every `Cell` in the `Triangle` could be
    accident-basis, gross of reinsurance, including DCC. This could be supplied by a
    single `Metadata` object in the function call instead of adding several constant columns
    to the `DataFrame`.

    Args:
        df: Pandas `DataFrame` to convert to a `Triangle`.
        loss_detail_cols: List of columns that belong to `loss_details` in `Metadata`, rather
            than `details`.
        metadata: Default metadata to apply to every cell in the `Triangle`.
    """
    _check_index_columns(df)
    if "field" not in df:
        raise Exception(
            "Long DataFrame must have a column `field` to convert to a Triangle"
        )
    if not pd.api.types.is_string_dtype(df.field):
        raise Exception("Long DataFrame field column must be a string")
    if "value" not in df:
        raise Exception(
            "Long DataFrame must have a column `value` to convert to a Triangle"
        )
    if not pd.api.types.is_numeric_dtype(df.value):
        raise Exception("Long DataFrame value column must be numeric")

    if loss_detail_cols is None:
        loss_detail_cols = []
    detail_cols = list(
        set(df.columns.values) - CORE_SET - {"field", "value"} - set(loss_detail_cols)
    )

    # noinspection DuplicatedCode
    if metadata is None:
        metadata = Metadata()
    cell_metadata = _create_metadata(df, metadata, detail_cols, loss_detail_cols)
    period_starts = list(df.period_start.dt.date)
    period_ends = list(df.period_end.dt.date)
    evaluation_dates = list(df.evaluation_date.dt.date)
    if "prev_evaluation_date" in df.columns:
        prev_evaluation_dates = list(df.prev_evaluation_date.dt.date)
    fields = list(df.field)
    values = list(df.value)

    cells = {}
    if "prev_evaluation_date" in df.columns:
        for (
            metadata,
            period_start,
            period_end,
            evaluation_date,
            prev_evaluation_date,
            field,
            value,
        ) in zip(
            cell_metadata,
            period_starts,
            period_ends,
            evaluation_dates,
            prev_evaluation_dates,
            fields,
            values,
        ):
            index = (
                period_start,
                period_end,
                evaluation_date,
                prev_evaluation_date,
                metadata,
            )
            if value is not None:
                if index in cells:
                    if field in cells[index].values:
                        raise Exception(
                            f"Field {field} is already present in this `Cell`"
                        )
                    cells[index].values[field] = np.array(value)
                else:
                    cells[index] = IncrementalCell(
                        period_start=period_start,
                        period_end=period_end,
                        evaluation_date=evaluation_date,
                        prev_evaluation_date=prev_evaluation_date,
                        metadata=metadata,
                        values={field: np.array(value)},
                    )
    else:
        if "risk_basis" not in df.columns:
            df = df.assign(risk_basis=[cell.risk_basis for cell in cell_metadata])
        if "per_occurrence_limit" not in df.columns:
            df = df.assign(
                per_occurrence_limit=[
                    cell.per_occurrence_limit for cell in cell_metadata
                ]
            )
        for idx, group_df in df.groupby(
            [
                "period_start",
                "period_end",
                "evaluation_date",
                "field",
                "risk_basis",
                "per_occurrence_limit",
            ]
            + detail_cols
            + loss_detail_cols,
            dropna=False,
        ):
            period_start, period_end, evaluation_date, field, *_ = idx
            current_metadata = _create_metadata(
                group_df, metadata, detail_cols, loss_detail_cols
            )[0]
            index = (
                period_start,
                period_end,
                evaluation_date,
                current_metadata,
            )
            if len(group_df) > 1:
                try:
                    group_df = group_df.sort_values(["scenario"])
                except KeyError:
                    raise Exception(
                        f"Field {field} is already present in this Cell. If multiple values are present for a single field they must be identified by a `scenario` column."
                    )
            value = group_df["value"].values
            # Check for scalar values, typically earned_premium
            if len(group_df) == 1 or (
                "scenario" in group_df and all(pd.isnull(group_df["scenario"]))
            ):
                value = value[0]
            if value is not None:
                if index in cells:
                    if field in cells[index].values:
                        raise Exception(
                            f"Field {field} is already present in this `Cell`"
                        )
                    cells[index].values[field] = value
                else:
                    cells[index] = CumulativeCell(
                        period_start=period_start.date(),
                        period_end=period_end.date(),
                        evaluation_date=evaluation_date.date(),
                        metadata=current_metadata,
                        values={field: value},
                    )
    return Triangle([cell for cell in cells.values()])


def _check_index_columns(df: pd.DataFrame) -> None:
    for column in INDEX_CUM_COLUMNS:
        if column not in df:
            raise Exception(
                f"DataFrame must have a column `{column}` to convert to a Triangle"
            )
        if not pd.api.types.is_datetime64_any_dtype(df[column]):
            if df[column].map(lambda i: isinstance(i, datetime.date)).all():
                df[column] = pd.to_datetime(df[column])
            else:
                raise Exception(f"DataFrame column {column} must be a datetime")
    if "prev_evaluation_date" in df.columns:
        if not pd.api.types.is_datetime64_any_dtype(df["prev_evaluation_date"]):
            raise Exception(
                "DataFrame column `prev_evaluation_date` must be a datetime"
            )


def _create_metadata(
    df: pd.DataFrame,
    metadata: Metadata,
    detail_columns: list[str],
    loss_detail_columns: list[str],
) -> list[Metadata]:
    risk_bases = _clean_metadata_column(df, "risk_basis", metadata.risk_basis)
    countries = _clean_metadata_column(df, "country", metadata.country)
    currencies = _clean_metadata_column(df, "currency", metadata.currency)
    reinsurance_bases = _clean_metadata_column(
        df, "reinsurance_basis", metadata.reinsurance_basis
    )
    loss_definitions = _clean_metadata_column(
        df, "loss_definition", metadata.loss_definition
    )
    per_occurrence_limits = _clean_metadata_column(
        df, "per_occurrence_limit", metadata.per_occurrence_limit
    )

    # Build details and loss_details columns
    pure_detail_columns = list(set(detail_columns) - set(loss_detail_columns))
    details_list = _create_metadata_details(df, pure_detail_columns, metadata.details)
    loss_details_list = _create_metadata_details(
        df, loss_detail_columns, metadata.loss_details
    )

    return [
        Metadata(
            risk_basis=risk_basis,
            country=country,
            currency=currency,
            reinsurance_basis=reinsurance_basis,
            loss_definition=loss_definition,
            per_occurrence_limit=per_occurrence_limit,
            details=details,
            loss_details=loss_details,
        )
        for (
            risk_basis,
            country,
            currency,
            reinsurance_basis,
            loss_definition,
            per_occurrence_limit,
            details,
            loss_details,
        ) in zip(
            risk_bases,
            countries,
            currencies,
            reinsurance_bases,
            loss_definitions,
            per_occurrence_limits,
            details_list,
            loss_details_list,
        )
    ]


def _create_metadata_details(
    df: pd.DataFrame,
    detail_columns: list[str],
    default_details: dict[str, Any],
) -> list[dict[str, Any]]:
    # Load up the details from the dataframe
    details = [{} for _ in range(df.shape[0])]
    for colname in detail_columns:
        clean_column = _clean_list_column(df[colname], default_details.get(colname))
        for i, val in enumerate(clean_column):
            if val is not None:
                details[i][colname] = val

    # Load in columns in the defaults but not in the dataframe
    default_cols = {col for col in default_details if col not in detail_columns}
    for colname in default_cols:
        for i in range(len(details)):
            details[i][colname] = default_details[colname]

    return details


def _clean_metadata_column(df: pd.DataFrame, colname: str, default: Any) -> list[Any]:
    if colname in df:
        return _clean_list_column(df[colname], default)
    else:
        return [default for _ in range(df.shape[0])]


def _clean_list_column(column: pd.Series, default: Any = None) -> list[Any]:
    return [default if missing else e for e, missing in zip(column, column.isna())]



from ..date_utils import add_months
from ..triangle import Triangle


def backfill(
    triangle: Triangle,
    static_fields: list[str] = ["earned_premium"],
    eval_resolution: int | None = None,
    min_dev_lag: int = 0,
) -> Triangle:
    """Backfill missing cells for each period with $0 loss, 0 claim counts
    and the first value of earned premium.

    Args:
        triangle: Triangle with missing cells.
        static_fields: Fields for which values will be preseved in the backfill.
        eval_resolution: Resolution of the evaluation date. If None, it will be
            calculated from the triangle.
        min_dev_lag: Minimum development lag for the backfilled cells. Can be negative,
            but must be greater than -period_resolution.
    """
    if eval_resolution is None:
        eval_resolution = triangle.eval_date_resolution
    additional_cells = []
    for _, period in triangle.period_rows:
        first_cell = period[0]
        replacement_values = {k: 0 for k in first_cell.values.keys()}
        for field in static_fields:
            replacement_values[field] = first_cell.values[field]
        current_lag = first_cell.dev_lag()
        while (current_lag - eval_resolution) >= min_dev_lag:
            current_lag -= eval_resolution
            additional_cells.append(
                first_cell.replace(
                    evaluation_date=lambda cell: add_months(
                        cell.period_end, current_lag
                    ),
                    values=replacement_values.copy(),
                )
            )

    return triangle + Triangle(additional_cells)



from dataclasses import dataclass, field

import numpy as np

from .index import MatrixIndex

__all__ = [
    "MissingValue",
    "DisaggregatedValue",
    "PredictedValue",
    "DisaggregatedPredictedValue",
    "Matrix",
    "RichMatrix",
]


@dataclass
class MissingValue(object):
    id: int


@dataclass(frozen=True)
class DisaggregatedValue(object):
    id: int
    value: float


@dataclass(frozen=True)
class PredictedValue(object):
    value: np.ndarray | None = field(default=None)


@dataclass(frozen=True)
class DisaggregatedPredictedValue(object):
    id: int
    value: np.ndarray | None = field(default=None)


class Matrix(object):
    def __init__(self, data: np.ndarray, index: MatrixIndex, incremental: bool = False):
        self.data = data
        self.index = index
        self.incremental = incremental

    def subset(self, args) -> "Matrix":
        int_slices, int_fields, int_exps, int_devs = self.index.resolve_indices(*args)
        return Matrix(
            data=self.data[int_slices, int_fields, int_exps, int_devs],
            index=self.index.subset(
                self.data.shape, int_slices, int_fields, int_exps, int_devs
            ),
        )

    def __getitem__(self, args) -> np.ndarray:
        int_slices, int_fields, int_exps, int_devs = self.index.resolve_indices(*args)
        return self.data[int_slices, int_fields, int_exps, int_devs]

    def __setitem__(self, args, value):
        int_slices, int_fields, int_exps, int_devs = self.index.resolve_indices(*args)
        self.data[int_slices, int_fields, int_exps, int_devs] = value


class RichMatrix(Matrix):
    pass



from __future__ import annotations

import altair as alt
from babel.numbers import get_currency_symbol
from dataclasses import dataclass
from functools import cache, wraps
from frozendict import frozendict
import json
import pandas as pd
import numpy as np
import string
from typing import Callable, Any, Literal

from .triangle import Triangle, Cell
from .base import metadata_diff
from .date_utils import period_resolution, eval_date_resolution

alt.renderers.enable("browser")

SLICE_TITLE_KWARGS = {
    "anchor": "middle",
    "font": "sans-serif",
    "fontWeight": "normal",
    "fontSize": 12,
}

BASE_HEIGHT = 600
BASE_WIDTH = "container"

BASE_AXIS_LABEL_FONT_SIZE = 16
BASE_AXIS_TITLE_FONT_SIZE = 18
FONT_SIZE_DECAY_FACTOR = 0.2

CellArgs = Cell | Cell, Cell, Cell
MetricFunc = Callable[[CellArgs], float | int | np.ndarray]
MetricFuncDict = dict[str, MetricFunc]

COMMON_METRIC_DICT: MetricFuncDict = {
    "Paid Loss Ratio": lambda cell: 100 * cell["paid_loss"] / cell["earned_premium"],
    "Reported Loss Ratio": lambda cell: 100
    * cell["reported_loss"]
    / cell["earned_premium"],
    "Incurred Loss Ratio": lambda cell: 100
    * cell["incurred_loss"]
    / cell["earned_premium"],
    "Paid Loss": lambda cell: cell["paid_loss"],
    "Reported Loss": lambda cell: cell["reported_loss"],
    "Incurred Loss": lambda cell: cell["incurred_loss"],
    "Earned Premium": lambda cell: cell["earned_premium"],
    "Reported Claims": lambda cell: cell["reported_claims"],
    "Paid ATA": lambda cell, prev_cell: cell["paid_loss"] / prev_cell["paid_loss"],
    "Reported ATA": lambda cell, prev_cell: cell["reported_loss"]
    / prev_cell["reported_loss"],
    "Paid Incremental ATA": lambda cell, prev_cell: cell["paid_loss"]
    / prev_cell["paid_loss"]
    - 1,
    "Reported Incremental ATA": lambda cell, prev_cell: cell["reported_loss"]
    / prev_cell["reported_loss"]
    - 1,
}

MetricFuncSpec = MetricFuncDict | str | list[str]


@dataclass
class FieldSummary(object):
    field: str
    metric: float | np.ndarray | None
    mean: float | None = None
    median: float | None = None
    sd: float | None = None
    min: float | None = None
    max: float | None = None
    q2_5: float | None = None
    q5: float | None = None
    q10: float | None = None
    q20: float | None = None
    q50: float | None = None
    q80: float | None = None
    q90: float | None = None
    q95: float | None = None
    q97_5: float | None = None
    is_forecast: bool = False
    keep_samples: bool = False

    def __post_init__(self):
        if self.metric is not None:
            if self.keep_samples:
                self.metric = {i: v for i, v in enumerate(self.metric)}
            else:
                self.metric = np.mean(self.metric)
            if self.mean is None:
                self.mean = np.mean(self.metric)

        if self.sd:
            self.is_forecast = True

    def tooltip(self, unit: str = "") -> str:
        mean = f"{self.mean:,.{self.precision}f}"
        sd = "" if self.sd is None else f" (SD: {self.sd:,.{self.precision}f})"
        if unit:
            return f"{self.field} ({unit}): {mean}{sd}"
        return f"{self.field}: {mean}{sd}"

    @property
    def precision(self) -> str:
        return "0" if self.mean > 100 else "2"

    @property
    def snake_case_field(self):
        return _to_snake_case(self.field)

    @staticmethod
    def quantiles() -> list[float]:
        return [0.25, 0.05, 0.1, 0.2, 0.5, 0.6, 0.9, 0.95, 0.975]

    @classmethod
    def from_metric(
        cls, name: str, metric: np.ndarray, keep_samples: bool = False
    ) -> FieldSummary:
        return cls(
            name,
            metric,
            np.mean(metric),
            np.median(metric),
            np.std(metric),
            np.min(metric),
            np.max(metric),
            *np.quantile(metric, cls.quantiles()),
            keep_samples=keep_samples,
        )

    def dict(self, return_empty: bool = False, unit: str = ""):
        if self.mean is None:
            return {}
        return {
            **self.__dict__,
            "tooltip": self.tooltip(unit),
            "snake_case_field": self.snake_case_field,
            "unit": unit,
            "is_forecast": self.is_forecast,
        }

    def json(self, return_empty: bool = False, unit: str = ""):
        return json.dumps(
            {
                k: "null" if v is None else v
                for k, v in self.dict(return_empty, unit).items()
            }
        )


def _resolve_metric_spec(metric_spec: MetricFuncSpec) -> MetricFuncDict:
    if isinstance(metric_spec, str):
        metric_spec = [metric_spec]
    if isinstance(metric_spec, list):
        result = {}
        for ref in metric_spec:
            if not isinstance(ref, str):
                raise ValueError("Supplied metric references must be strings")
            elif ref not in COMMON_METRIC_DICT:
                raise ValueError(f"Don't know the definition of metric {ref}")
            else:
                result[ref] = COMMON_METRIC_DICT[ref]
        return result
    else:
        return metric_spec


# Pattern from https://stackoverflow.com/a/53394430
def freezeargs(func):
    """Convert a mutable dictionary into immutable.
    Useful to be compatible with cache
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        args = (frozendict(arg) if isinstance(arg, dict) else arg for arg in args)
        kwargs = {
            k: frozendict(v) if isinstance(v, dict) else v for k, v in kwargs.items()
        }
        return func(*args, **kwargs)

    return wrapped


def _flatten_dict(x: dict, prepend: str = "") -> dict[str, Any]:
    out = {}
    for key, value in x.items():
        if isinstance(value, dict):
            out |= _flatten_dict(value, prepend=str(key) + "_")
        else:
            out[prepend + str(key)] = value
    return out


def _to_snake_case(x: str) -> str:
    return x.replace(" ", "_").lower()


@freezeargs
@cache
def build_plot_data(
    triangle: Triangle,
    metric_dict: MetricFuncDict | None = None,
    remove_empties: bool = True,
    flat: bool = False,
    keep_samples: bool = False,
) -> list[dict[str, Any]]:
    """Convenience function for building plot data.

    By default, this function will calculate all plot data
    for metrics in COMMON_METRIC_DICT. If bespoke metrics
    are needed, users can pass in a `metric_dict` of their own.

    The returned object is a list of dictionaries. If `flat=False`,
    then each dictionary is nested, with each metric having it's own
    metric spec, e.g. [{'paid_loss_ratio': {'mean': 0.25, 'sd': 0.05}}].
    If `flat=True`, then the dictionaries are non-nested,
    e.g. [{'paid_loss_ratio_mean': 0.25, 'paid_loss_ratio_sd': 0.05}}].

    Args:
        triangle: The triangle to plot.
        metric_dict: A MetricFuncDict to override the default metrics used.
        remove_empties: Remove any cells that don't have data.
        flat: Return a flat dictionary.
        keep_samples: Keep field samples (e.g. distribution variates) in the summary.
    """
    currency = _currency_symbol(triangle)
    unit_lookup = {
        field: (
            currency
            if ("loss" in field.lower() or "premium" in field.lower())
            and "ratio" not in field.lower()
            else "%"
            if "ratio" in field.lower()
            else "%"
            if "share" in field.lower()
            else "N"
            if "claims" in field.lower()
            else ""
        )
        for field in (metric_dict or COMMON_METRIC_DICT)
    }
    field_summaries = {
        cell: {
            _to_snake_case(name): _calculate_field_summary(
                cell, prev_cell, metric, name, keep_samples=keep_samples
            ).dict(remove_empties, unit_lookup.get(name, ""))
            for name, metric in (metric_dict or COMMON_METRIC_DICT).items()
        }
        for cell, prev_cell in zip(triangle, [None, *triangle[:-1]])
    }

    if remove_empties:
        field_summaries = {
            cell: {name: summary for name, summary in values.items() if summary}
            for cell, values in field_summaries.items()
        }

    plot_data = [
        {
            **_core_plot_data(cell),
            "last_lag": max(
                triangle.filter(lambda ob: ob.period == cell.period).dev_lags()
            ),
            "last_observed_lag": max(
                _remove_triangle_samples(triangle)
                .filter(lambda ob: ob.period == cell.period)
                .dev_lags()
                or [0]
            )
            or None,
            "fields": list(cell.values),
            "exp_resolution": period_resolution(triangle),
            "eval_resolution": eval_date_resolution(triangle),
            "tooltip": ", ".join(
                [
                    v["tooltip"]
                    for v in field_summaries[cell].values()
                    if v["snake_case_field"] in cell.values
                ]
            ),
            **field_summaries[cell],
        }
        for cell in triangle
    ]
    if flat:
        return [_flatten_dict(v) for v in plot_data]
    return plot_data


@alt.theme.register("bermuda_plot_theme", enable=True)
def bermuda_plot_theme() -> alt.theme.ThemeConfig:
    return {
        "autosize": {"contains": "content", "resize": True},
        "config": {
            "style": {
                "group-title": {"fontSize": 24},
                "group-subtitle": {"fontSize": 18},
                "guide-label": {
                    "fontSize": BASE_AXIS_LABEL_FONT_SIZE,
                    "font": "sans-serif",
                },
                "guide-title": {
                    "fontSize": BASE_AXIS_TITLE_FONT_SIZE,
                    "font": "sans-serif",
                },
            },
            "mark": {"color": "black"},
            "title": {"anchor": "start", "offset": 20},
            "axis": {"labelOverlap": True},
            "legend": {
                "orient": "right",
                "titleAnchor": "start",
                "layout": {
                    "direction": "vertical",
                },
            },
        },
    }


def _remove_triangle_samples(triangle: Triangle) -> Triangle:
    """Removes cells that contain samples. The primary use-case
    of this method is to remove future predictions from a triangle to make
    investigating observed data in combined triangles easier."""
    if triangle.num_samples == 1:
        return triangle

    int_cells = []
    for cell in triangle:
        if not any(
            isinstance(v, np.ndarray) and v.size > 1 for v in cell.values.values()
        ):
            int_cells.append(cell)
    return Triangle(int_cells)


def plot_right_edge(
    triangle: Triangle,
    hide_samples: bool = False,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    main_title = alt.Title(
        "Latest Loss Ratio", subtitle="The most recent loss ratio diagonal"
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        _remove_triangle_samples(triangle) if hide_samples else triangle,
        plot_func=_plot_right_edge,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        uncertainty_type=uncertainty_type,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_right_edge(
    triangle: Triangle,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
) -> alt.Chart:
    if "earned_premium" not in triangle.fields:
        raise ValueError(
            "Triangle must contain `earned_premium` to plot its right edge. "
            f"This triangle contains {triangle.fields}"
        )

    data = alt.Data(values=build_plot_data(triangle.right_edge, COMMON_METRIC_DICT))

    currency = _currency_symbol(triangle)

    bar = (
        alt.Chart(data, title=title)
        .mark_bar()
        .encode(
            x=alt.X("yearmonth(period_start):O"),
            y=alt.Y(f"earned_premium.mean:Q").axis(
                title="Earned Premium", format="$.2s"
            ),
            color=alt.Color("earned_premium.field:N")
            .scale(range=["lightgray"])
            .title("Field"),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("dev_lag:O", title="Dev Lag"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip(
                    f"earned_premium.tooltip:N",
                    title="Bar",
                ),
            ],
        )
    )

    loss_error = alt.LayerChart()
    loss_fields = sorted(
        set(
            [
                key
                for cell_values in data.values
                for key in cell_values
                if "loss_ratio" in key
            ]
        )
    )
    lines = alt.LayerChart()
    points = alt.LayerChart()

    for field in loss_fields:
        snake_name = _to_snake_case(field)
        if uncertainty and uncertainty_type == "ribbon":
            loss_error += (
                alt.Chart(data)
                .mark_area(
                    opacity=0.5,
                )
                .encode(
                    x=alt.X("yearmonth(period_start):T"),
                    y=alt.Y(f"{snake_name}.q5:Q").axis(
                        title="Loss Ratio %", format=".0f"
                    ),
                    y2=alt.Y2(f"{snake_name}.q95:Q"),
                    color=alt.Color(f"{snake_name}.field:N"),
                )
            )
        elif uncertainty and uncertainty_type == "segments":
            loss_error += (
                alt.Chart(data)
                .mark_rule(thickness=3)
                .encode(
                    x=alt.X("yearmonth(period_start):T").title("Period Start"),
                    y=alt.Y(f"{snake_name}.q5:Q").axis(
                        title="Loss Ratio %", format=".0f"
                    ),
                    y2=alt.Y2(f"{snake_name}.q95:Q"),
                    color=alt.Color(f"{snake_name}.field:N"),
                )
            )

        lines += (
            alt.Chart(data)
            .mark_line(
                size=1,
            )
            .encode(
                x=alt.X("yearmonth(period_start):T", axis=alt.Axis(labelAngle=0)).title(
                    "Period Start"
                ),
                y=alt.Y(
                    f"{snake_name}.mean:Q",
                    scale=alt.Scale(zero=True),
                    axis=alt.Axis(format=".0f"),
                ).title("Loss Ratio %"),
                color=alt.Color(f"{snake_name}.field:N").legend(title=None),
            )
        ).interactive()

        points += (
            alt.Chart(data)
            .mark_point(
                size=max(20, 100 / mark_scaler),
                filled=True,
                opacity=1,
            )
            .encode(
                x=alt.X("yearmonth(period_start):T", axis=alt.Axis(labelAngle=0)).title(
                    "Period Start"
                ),
                y=alt.Y(
                    f"{snake_name}.mean:Q",
                    scale=alt.Scale(zero=True),
                    axis=alt.Axis(format=".0f"),
                ).title("Loss Ratio %"),
                color=alt.Color(f"{field}.field:N").legend(title=None),
                tooltip=[
                    alt.Tooltip("period_start:T", title="Period Start"),
                    alt.Tooltip("period_end:T", title="Period End"),
                    alt.Tooltip("dev_lag:O", title="Dev Lag"),
                    alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                    alt.Tooltip(f"{snake_name}.tooltip:N", title="Field"),
                ],
            )
        )

    fig = alt.layer(bar, loss_error + lines + points).resolve_scale(
        y="independent",
        color="independent",
    )

    return fig.interactive()


def plot_data_completeness(
    triangle: Triangle,
    hide_samples: bool = False,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    main_title = alt.Title(
        "Triangle Completeness",
        subtitle="The number of data fields available per cell",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        _remove_triangle_samples(triangle) if hide_samples else triangle,
        plot_func=_plot_data_completeness,
        title=main_title,
        facet_titles=facet_titles,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_data_completeness(
    triangle: Triangle, title: alt.Title, mark_scaler: int
) -> alt.Chart:
    if not triangle.is_disjoint:
        raise Exception(
            "This triangle isn't disjoint! You probably don't want to use it"
        )
    if not triangle.is_semi_regular:
        raise Exception(
            "This triangle isn't semi-regular! You probably don't want to use it"
        )

    currency = _currency_symbol(triangle)

    selection = alt.selection_point()

    data = alt.Data(values=build_plot_data(triangle))

    fig = (
        alt.Chart(
            data,
            title=title,
        )
        .transform_calculate(n_fields=alt.expr.length(alt.datum.fields))
        .mark_circle(size=500 * 1 / mark_scaler, opacity=1)
        .encode(
            alt.X(
                "dev_lag:N", axis=alt.Axis(labelAngle=0), scale=alt.Scale(zero=True)
            ).title("Dev Lag (months)"),
            alt.Y(
                "yearmonth(period_start):T", scale=alt.Scale(padding=15, reverse=True)
            ).title("Period Start"),
            color=alt.condition(
                selection,
                alt.Color("n_fields:N").scale(scheme="dark2").title("Number of Fields"),
                alt.value("lightgray"),
            ),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("dev_lag:N", title="Dev Lag (months)"),
                alt.Tooltip("tooltip:N", title="Fields"),
            ],
        )
        .add_params(selection)
    )

    return fig.interactive()


def plot_heatmap(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    hide_samples: bool = False,
    show_values: bool = True,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a heatmap."""
    main_title = alt.Title(
        "Triangle Heatmap",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_heatmap,
            metric_dict=metric_dict,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            mark_scaler=max_cols,
            ncols=max_cols,
            show_values=show_values,
        )
        .resolve_scale(color="independent")
        .resolve_legend(color="independent")
    )
    return fig


def _plot_heatmap(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    show_values: bool,
    title: alt.Title,
    mark_scaler: int,
) -> alt.Chart:
    snake_name = _to_snake_case(name)
    data = alt.Data(values=build_plot_data(triangle, {name: metric}))

    base = alt.Chart(data, title=title).encode(
        x=alt.X("dev_lag:N", axis=alt.Axis(labelAngle=0)).title("Dev Lag (months)"),
        y=alt.X("yearmonth(period_start):O", scale=alt.Scale(reverse=False)).title(
            "Period Start"
        ),
    )

    selection = alt.selection_interval()
    heatmap = (
        base.mark_rect()
        .encode(
            color=alt.when(selection)
            .then(
                alt.Color(
                    f"{snake_name}.mean:Q",
                    scale=alt.Scale(scheme="blueorange"),
                    legend=alt.Legend(title=name, format=".2s"),
                ).title(name)
            )
            .otherwise(
                alt.value("gray"),
            ),
            tooltip=[
                alt.Tooltip("period_start:T", title="Period Start"),
                alt.Tooltip("period_end:T", title="Period End"),
                alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
                alt.Tooltip("dev_lag:O", title="Dev Lag (months)"),
                alt.Tooltip(f"{snake_name}.tooltip:N", title="Field"),
            ],
        )
        .add_params(selection)
    )

    if show_values:
        text = base.mark_text(
            fontSize=BASE_AXIS_TITLE_FONT_SIZE
            * np.exp(-FONT_SIZE_DECAY_FACTOR * mark_scaler),
            font="monospace",
        ).encode(text=alt.Text(f"{snake_name}.mean:Q", format=".0f"))

        return heatmap + text
    return heatmap.resolve_scale(color="independent")


def plot_atas(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid ATA"],
    hide_samples: bool = False,
    ncols: int | None = None,
    width: int = 400,
    height: int = 200,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle ATAs."""
    main_title = alt.Title(
        "Triangle ATAs",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_atas,
            metric_dict=metric_dict,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
        .resolve_scale(color="independent")
    )
    return fig


def _plot_atas(
    triangle: Triangle, metric: MetricFunc, name: str, title: alt.Title
) -> alt.Chart:
    data = alt.Data(values=build_plot_data(triangle, {name: metric}, flat=True))
    snake_name = _to_snake_case(name)

    tooltip = [
        alt.Tooltip("period_start:T", title="Period Start"),
        alt.Tooltip("period_end:T", title="Period End"),
        alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
        alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
        alt.Tooltip(f"{snake_name}_tooltip:N", title="Field"),
    ]

    base = alt.Chart(data, title=title).encode(
        x=alt.X("dev_lag:Q").title("Dev Lag (months)").scale(padding=10),
        y=alt.Y(f"{snake_name}_mean:Q").title(name).scale(zero=False, padding=10),
        tooltip=tooltip,
    )

    points = base.mark_point(color="black", filled=True)
    boxplot = base.mark_boxplot(
        opacity=0.7,
        color="skyblue",
        median=alt.MarkConfig(stroke="black"),
        rule=alt.MarkConfig(stroke="black"),
        box=alt.MarkConfig(stroke="black"),
    )
    errors = base.mark_rule(thickness=1).encode(
        y=alt.Y(f"{snake_name}_:Q").axis(title=name),
        y2=alt.Y2(f"{snake_name}_q95:Q"),
        color=alt.value("black"),
        opacity=alt.value(0.7),
    )

    return (points + errors + boxplot).interactive()


def plot_growth_curve(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    hide_samples: bool = False,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments", "spaghetti"] = "ribbon",
    n_lines: int = 100,
    seed: int | None = None,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a growth curve."""
    main_title = alt.Title(
        "Triangle Growth Curve",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_growth_curve,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            n_lines=n_lines,
            seed=seed,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
    )
    return fig


def _plot_growth_curve(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments", "spaghetti"],
    mark_scaler: int,
    n_lines: int = 100,
    seed: int | None = None,
) -> alt.Chart:
    snake_name = _to_snake_case(name)

    if uncertainty_type == "spaghetti":
        triangle_thinned = triangle.thin(num_samples=n_lines)
        data = alt.Data(
            values=build_plot_data(
                triangle_thinned, metric_dict={name: metric}, keep_samples=True
            )
        )
    else:
        data = alt.Data(values=build_plot_data(triangle, metric_dict={name: metric}))

    period_start_dtype = "Q" if len(triangle.periods) > 10 else "N"
    color = (
        alt.Color(f"yearmonth(period_start):{period_start_dtype}")
        .scale(scheme="blueorange", reverse=True)
        .legend(title="Period Start")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )

    base = alt.Chart(data, title=title).encode(
        x=alt.X(
            "dev_lag:Q",
            axis=alt.Axis(grid=True, labelAngle=0),
            scale=alt.Scale(padding=5),
        ).title("Dev Lag (months)"),
        y=alt.Y(f"{snake_name}.mean:Q", axis=alt.Axis(format=".2s")).title(name),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{snake_name}.tooltip:N", title="Field"),
        ],
    )

    lines = base.mark_line(opacity=0.2).encode(color=color_conditional_no_legend)

    points = base.mark_point(stroke=None, filled=True).encode(
        color=color_conditional_no_legend,
        opacity=opacity_conditional,
    )

    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(
            color=color_conditional,
            opacity=opacity_conditional,
            strokeOpacity=opacity_conditional,
        )
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_opacity_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area(
            opacity=0.5,
        ).encode(
            y=alt.Y(f"{snake_name}.q5:Q"),
            y2=alt.Y2(f"{snake_name}.q95:Q"),
            color=color_conditional_no_legend,
            opacity=ribbon_opacity_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_rule(thickness=5).encode(
            y=alt.Y(f"{snake_name}.q5:Q").title(name),
            y2=alt.Y2(f"{snake_name}.q95:Q"),
            color=color_conditional_no_legend,
            opacity=opacity_conditional,
        )
    elif uncertainty and uncertainty_type == "spaghetti":
        errors = alt.LayerChart()
        for i in range(n_lines):
            errors += (
                alt.Chart(data)
                .mark_line(opacity=0.2)
                .encode(
                    x=alt.X("dev_lag:Q"),
                    y=alt.Y(f"{snake_name}.metric[{i}]:Q"),
                    detail=f"{i}:N",
                    color=color_conditional_no_legend,
                )
            )
    else:
        errors = alt.LayerChart()

    if len(triangle.periods) == 1:
        scale_color = "shared"
    else:
        scale_color = "independent"

    return (
        alt.layer(errors + lines + points, ultimates.add_params(selector))
        .resolve_scale(color=scale_color)
        .interactive()
    )


def plot_sunset(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Incremental ATA"],
    hide_samples: bool = False,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    """Plot triangle metrics as a sunset."""
    main_title = alt.Title(
        "Triangle Sunset",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_sunset,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols))
    )
    return fig.interactive()


def _plot_sunset(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments"],
) -> alt.Chart:
    snake_name = _to_snake_case(name)
    data = alt.Data(values=build_plot_data(triangle, {name: metric}, flat=True))

    color = (
        alt.Color("dev_lag:Q")
        .scale(scheme="blueorange")
        .legend(title="Development Lag")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["dev_lag"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(data, title=title).encode(
        x=alt.X(
            "yearmonth(evaluation_date):O", axis=alt.Axis(grid=True, labelAngle=0)
        ).title("Calendar Year"),
        y=alt.X(f"{snake_name}_mean:Q").title(name).scale(type="pow", exponent=0.3),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{snake_name}_tooltip:N", title="Field"),
        ],
    )

    points = base.mark_point(stroke=None, size=30 / mark_scaler, filled=True).encode(
        color=color_conditional,
        opacity=opacity_conditional,
        strokeOpacity=opacity_conditional,
    )
    regression = (
        base.transform_loess(
            "evaluation_date", f"{snake_name}_mean", groupby=["dev_lag"], bandwidth=0.6
        )
        .mark_line(strokeWidth=2)
        .encode(color=color_conditional_no_legend, opacity=opacity_conditional)
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area().encode(
            y=alt.Y(f"{snake_name}_q5:Q").axis(title=name),
            y2=alt.Y2(f"{snake_name}_q95:Q"),
            color=color_conditional_no_legend,
            opacity=ribbon_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_rule(thickness=5).encode(
            y=alt.Y(f"{snake_name}_q5:Q").axis(title=name),
            y2=alt.Y2(f"{snake_name}_q95:Q"),
            color=color_conditional_no_legend,
            opacity=opacity_conditional,
        )
    else:
        errors = alt.LayerChart()

    return (
        alt.layer(errors, regression, points)
        .add_params(selector)
        .resolve_scale(color="independent")
    )


def plot_mountain(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = ["Paid Loss Ratio"],
    hide_samples: bool = False,
    uncertainty: bool = True,
    uncertainty_type: Literal["ribbon", "segments"] = "ribbon",
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
    highlight_ultimates: bool = True,
) -> alt.Chart:
    """Plot triangle metrics as a mountain."""
    main_title = alt.Title(
        "Triangle Mountain Plot",
    )
    metric_dict = _resolve_metric_spec(metric_spec)
    n_metrics = len(metric_dict)
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    legend_direction = "horizontal" if highlight_ultimates else "vertical"
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_mountain,
            metric_dict=metric_dict,
            uncertainty=uncertainty,
            uncertainty_type=uncertainty_type,
            highlight_ultimates=highlight_ultimates,
            mark_scaler=max_cols,
            title=main_title,
            facet_titles=facet_titles,
            width=width,
            height=height,
            ncols=max_cols,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .configure_legend(**_compute_font_sizes(max_cols), direction=legend_direction)
    )
    return fig.interactive()


def _plot_mountain(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    uncertainty_type: Literal["ribbon", "segments"],
    highlight_ultimates: bool = True,
) -> alt.Chart:
    snake_name = _to_snake_case(name)

    if highlight_ultimates:
        metric_triangle = _remove_triangle_samples(triangle)
        data = alt.Data(values=build_plot_data(metric_triangle, {name: metric}))
        ultimate_data = alt.Data(
            values=build_plot_data(triangle.right_edge, {name: metric})
        )
    else:
        data = alt.Data(values=build_plot_data(triangle, {name: metric}))

    color = (
        alt.Color("dev_lag:Q")
        .scale(scheme="blueorange")
        .legend(title="Development Lag")
    )
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["dev_lag"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )

    tooltip = [
        alt.Tooltip("period_start:T", title="Period Start"),
        alt.Tooltip("period_end:T", title="Period End"),
        alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
        alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
        alt.Tooltip(f"{snake_name}.tooltip:N", title="Field"),
    ]

    base = alt.Chart(data, title=title).encode(
        x=alt.X(
            "yearmonth(period_start):O", axis=alt.Axis(grid=True, labelAngle=0)
        ).title("Period Start"),
        y=alt.Y(f"{snake_name}.mean:Q", axis=alt.Axis(format=".2s")).title(name),
        tooltip=tooltip,
    )

    lines = base.mark_line().encode(color=color_none, opacity=opacity_conditional)
    points = base.mark_point(filled=True, stroke=None).encode(
        color=color,
        opacity=opacity_conditional,
    )

    if uncertainty and uncertainty_type == "ribbon":
        ribbon_conditional = (
            alt.when(selector)
            .then(alt.OpacityValue(0.5))
            .otherwise(alt.OpacityValue(0.2))
        )
        errors = base.mark_area().encode(
            y=alt.Y(f"{snake_name}.q5:Q"),
            y2=alt.Y2(f"{snake_name}.q95:Q"),
            color=color_none,
            opacity=ribbon_conditional,
        )
    elif uncertainty and uncertainty_type == "segments":
        errors = base.mark_errorbar(thickness=5).encode(
            y=alt.Y(f"{snake_name}.q5:Q").title(name),
            y2=alt.Y2(f"{snake_name}.q95:Q"),
            color=color_none,
            opacity=opacity_conditional,
        )
    else:
        errors = alt.LayerChart()

    if highlight_ultimates:
        ultimate_color = alt.Color("dev_lag:Q").scale(scheme="greys")
        ultimate_base = alt.Chart(ultimate_data).encode(
            x=alt.X(
                "yearmonth(period_start):O", axis=alt.Axis(grid=True, labelAngle=0)
            ).title("Period Start"),
            y=alt.X(f"{snake_name}.mean:Q").title(name),
            tooltip=tooltip,
            order=alt.value(1),
        )
        ultimates = ultimate_base.mark_line().encode(
            color=ultimate_color.legend(title="Ultimate Lag")
        )
        ultimates += ultimate_base.mark_point(filled=True, size=100).encode(
            color=ultimate_color,
        )

        if uncertainty and uncertainty_type == "ribbon":
            ultimates += ultimate_base.mark_area().encode(
                y=alt.Y(f"{snake_name}.q5:Q"),
                y2=alt.Y2(f"{snake_name}.q95:Q"),
                color=ultimate_color,
                opacity=ribbon_conditional,
            )
        if uncertainty and uncertainty_type == "segments":
            ultimates += ultimate_base.mark_rule(thickness=5).encode(
                y=alt.Y(f"{snake_name}.q5:Q").title(name),
                y2=alt.Y2(f"{snake_name}.q95:Q"),
                color=ultimate_color,
                opacity=opacity_conditional,
            )
    else:
        ultimates = alt.LayerChart()

    return alt.layer(
        lines + errors,
        points.add_params(selector),
        ultimates,
    ).resolve_scale(color="independent")


def plot_ballistic(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
        "Reported Loss Ratio": lambda cell: 100
        * cell["reported_loss"]
        / cell["earned_premium"],
    },
    hide_samples: bool = False,
    uncertainty: bool = True,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
    show_points: bool = True,
) -> alt.Chart:
    """Plot triangle metrics as a ballistic."""
    main_title = alt.Title(
        "Triangle Ballistic Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        _remove_triangle_samples(triangle) if hide_samples else triangle,
        plot_func=_plot_ballistic,
        axis_metrics=axis_metrics,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
        show_points=show_points,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_ballistic(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    show_points: bool,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())
    snake_name_x = _to_snake_case(name_x)
    snake_name_y = _to_snake_case(name_y)

    data = alt.Data(values=build_plot_data(triangle, axis_metrics))

    color_var = "dev_lag" if show_points else "yearmonth(period_start)"
    color_title = "Development Lags (months)" if show_points else "Period Start"
    if show_points:
        color_dtype = "Q" if len(triangle.dev_lags()) > 10 else "N"
    else:
        color_dtype = "Q" if len(triangle.periods) > 10 else "N"

    color = alt.Color(f"{color_var}:{color_dtype}").legend(title=color_title)
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(data, title=title).encode(
        x=alt.X(f"{snake_name_x}.mean:Q").title(name_x).axis(grid=True),
        y=alt.X(f"{snake_name_y}.mean:Q").title(name_y).axis(grid=True),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{snake_name_x}.tooltip:N", title="Field X"),
            alt.Tooltip(f"{snake_name_y}.tooltip:N", title="Field Y"),
        ],
    )

    diagonal = (
        alt.Chart(data)
        .mark_line(color="black", strokeDash=[5, 5])
        .encode(
            x=f"{snake_name_x}.mean:Q",
            y=f"{snake_name_x}.mean:Q",
        )
    )

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )
    if show_points:
        points = base.mark_point(
            filled=True, size=100 / mark_scaler, stroke=None
        ).encode(color=color_conditional, opacity=opacity_conditional)
    else:
        points = alt.LayerChart()

    ultimates = (
        base.mark_point(size=200 / mark_scaler, filled=True, stroke=None)
        .encode(color=color_conditional, opacity=opacity_conditional)
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = (
            base.mark_rule(thickness=5)
            .encode(
                y=alt.Y(f"{snake_name_y}.q5:Q").axis(title=name_y),
                y2=alt.Y2(f"{snake_name_y}.q95:Q"),
                color=alt.ColorValue("black"),
            )
            .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
        )
    else:
        errors = alt.LayerChart()

    return (
        alt.layer(diagonal, errors + lines, (points + ultimates).add_params(selector))
        .resolve_scale(color="independent")
        .interactive()
    )


def plot_broom(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid/Reported Ratio": lambda cell: cell["paid_loss"] / cell["reported_loss"],
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
    },
    hide_samples: bool = False,
    rule: int | None = 1,
    uncertainty: bool = True,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
    show_points: bool = True,
) -> alt.Chart:
    """Plot triangle metrics as a broom."""
    main_title = alt.Title(
        "Triangle Broom Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = _build_metric_slice_charts(
        _remove_triangle_samples(triangle) if hide_samples else triangle,
        plot_func=_plot_broom,
        axis_metrics=axis_metrics,
        title=main_title,
        facet_titles=facet_titles,
        uncertainty=uncertainty,
        rule=rule,
        width=width,
        height=height,
        mark_scaler=max_cols,
        ncols=max_cols,
        show_points=show_points,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig.interactive()


def _plot_broom(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    mark_scaler: int,
    uncertainty: bool,
    rule: int | None,
    show_points: bool,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())
    snake_name_x = _to_snake_case(name_x)
    snake_name_y = _to_snake_case(name_y)

    data = alt.Data(values=build_plot_data(triangle, axis_metrics))

    color_var = "dev_lag" if show_points else "yearmonth(period_start)"
    color_title = "Development Lags (months)" if show_points else "Period Start"
    if show_points:
        color_dtype = "Q" if len(triangle.dev_lags()) > 10 else "N"
    else:
        color_dtype = "Q" if len(triangle.periods) > 10 else "N"

    color = alt.Color(f"{color_var}:{color_dtype}").legend(title=color_title)
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(data, title=title).encode(
        x=alt.X(f"{snake_name_x}.mean:Q").scale(padding=10, nice=False).title(name_x),
        y=alt.Y(f"{snake_name_y}.mean:Q").title(name_y),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{snake_name_x}.tooltip:N", title="Field X"),
            alt.Tooltip(f"{snake_name_y}.tooltip:N", title="Field Y"),
        ],
    )

    wall = (
        alt.Chart().mark_rule(strokeDash=[12, 5], opacity=0.5, strokeWidth=2)
    ).encode()
    if rule is not None:
        wall = wall.encode(x=alt.datum(rule))

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )
    if show_points:
        points = base.mark_point(
            filled=True, size=100 / mark_scaler, stroke=None
        ).encode(
            color=color_conditional,
            opacity=opacity_conditional,
            strokeOpacity=opacity_conditional,
        )
    else:
        points = alt.LayerChart()

    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(
            color=color_conditional,
            opacity=opacity_conditional,
            strokeOpacity=opacity_conditional,
        )
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = (
            base.mark_rule(thickness=5)
            .encode(
                x=alt.X(f"{snake_name_x}.q5:Q").axis(title=name_x),
                x2=alt.X2(f"{snake_name_x}.q95:Q"),
                color=alt.ColorValue("black"),
            )
            .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
        )
    else:
        errors = alt.LayerChart()

    return alt.layer(
        errors + lines + wall, (points + ultimates).add_params(selector)
    ).resolve_scale(color="independent")


def plot_drip(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Reported Loss Ratio": lambda cell: 100
        * cell["reported_loss"]
        / cell["earned_premium"],
        "Open Claim Share": lambda cell: 100
        * cell["open_claims"]
        / cell["reported_claims"],
    },
    hide_samples: bool = False,
    uncertainty: bool = True,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
    show_points: bool = True,
) -> alt.Chart:
    """Plot triangle metrics as a drip."""
    main_title = alt.Title(
        "Triangle Drip Plot",
    )
    n_slices = len(triangle.slices)
    max_cols = ncols or _determine_facet_cols(n_slices)
    fig = (
        _build_metric_slice_charts(
            _remove_triangle_samples(triangle) if hide_samples else triangle,
            plot_func=_plot_drip,
            axis_metrics=axis_metrics,
            title=main_title,
            facet_titles=facet_titles,
            uncertainty=uncertainty,
            width=width,
            height=height,
            mark_scaler=max_cols,
            ncols=max_cols,
            show_points=show_points,
        )
        .configure_axis(**_compute_font_sizes(max_cols))
        .resolve_scale(color="independent")
    )
    return fig.interactive()


def _plot_drip(
    triangle: Triangle,
    axis_metrics: MetricFuncDict,
    title: alt.Title,
    uncertainty: bool,
    mark_scaler: int,
    show_points: bool,
) -> alt.Chart:
    (name_x, name_y), (func_x, func_y) = zip(*axis_metrics.items())
    snake_name_x = _to_snake_case(name_x)
    snake_name_y = _to_snake_case(name_y)

    data = alt.Data(values=build_plot_data(triangle, axis_metrics))

    color_var = "dev_lag" if show_points else "yearmonth(period_start)"
    color_title = "Development Lags (months)" if show_points else "Period Start"
    if show_points:
        color_dtype = "Q" if len(triangle.dev_lags()) > 10 else "N"
    else:
        color_dtype = "Q" if len(triangle.periods) > 10 else "N"

    color = alt.Color(f"{color_var}:{color_dtype}").legend(title=color_title)
    color_none = color.legend(None)

    selector = alt.selection_point(fields=["period_start"])
    opacity_conditional = (
        alt.when(selector).then(alt.OpacityValue(1)).otherwise(alt.OpacityValue(0.2))
    )
    color_conditional = alt.when(selector).then(color).otherwise(alt.value("lightgray"))
    color_conditional_no_legend = (
        alt.when(selector).then(color_none).otherwise(alt.value("lightgray"))
    )

    base = alt.Chart(data, title=title).encode(
        x=alt.X(f"{snake_name_x}.mean:Q").title(name_x, padding=10),
        y=alt.Y(f"{snake_name_y}.mean:Q").title(name_y).scale(nice=False, padding=10),
        tooltip=[
            alt.Tooltip("period_start:T", title="Period Start"),
            alt.Tooltip("period_end:T", title="Period End"),
            alt.Tooltip("evaluation_date:T", title="Evaluation Date"),
            alt.Tooltip("dev_lag:Q", title="Dev Lag (months)"),
            alt.Tooltip(f"{snake_name_x}.tooltip:N", title="Field X"),
            alt.Tooltip(f"{snake_name_y}.tooltip:N", title="Field Y"),
        ],
    )

    lines = base.mark_line(color="black", strokeWidth=0.5).encode(
        detail="period_start:N", opacity=opacity_conditional
    )

    if show_points:
        points = base.mark_point(
            filled=True, size=100 / mark_scaler, stroke=None
        ).encode(color=color_conditional, opacity=opacity_conditional)
    else:
        points = alt.LayerChart()

    ultimates = (
        base.mark_point(size=300 / mark_scaler, filled=True, stroke=None)
        .encode(color=color_conditional, opacity=opacity_conditional)
        .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
    )

    if uncertainty:
        errors = (
            base.mark_rule(thickness=5)
            .encode(
                y=alt.Y(f"{snake_name_y}.q5:Q").title(name_y),
                y2=alt.Y2(f"{snake_name_y}.q95:Q"),
                color=alt.ColorValue("black"),
            )
            .transform_filter(alt.datum.last_lag == alt.datum.dev_lag)
        )
    else:
        errors = alt.LayerChart()

    return alt.layer(
        errors + lines, (points + ultimates).add_params(selector)
    ).resolve_scale(color="independent")


def plot_hose(
    triangle: Triangle,
    axis_metrics: MetricFuncDict = {
        "Paid Loss Ratio": lambda cell: 100
        * cell["paid_loss"]
        / cell["earned_premium"],
        "Incremental Paid Loss Ratio": lambda cell, prev_cell: 100
        * (
            cell["paid_loss"] / cell["earned_premium"]
            - prev_cell["paid_loss"] / prev_cell["earned_premium"]
        ),
    },
    hide_samples: bool = False,
    uncertainty: bool = True,
    width: int = 400,
    height: int = 300,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
    show_points: bool = True,
) -> alt.Chart:
    return plot_drip(
        triangle,
        axis_metrics,
        hide_samples,
        uncertainty,
        width,
        height,
        ncols,
        facet_titles,
        show_points=show_points,
    ).properties(title="Triangle Hose Plot")


def plot_histogram(
    triangle: Triangle,
    metric_spec: MetricFuncSpec = "Paid Loss",
    right_edge: bool = True,
    hide_samples: bool = False,
    width: int = 400,
    height: int = 200,
    ncols: int | None = None,
    facet_titles: list[str] | None = None,
) -> alt.Chart:
    main_title = alt.Title("Triangle Histogram")
    metric_dict = _resolve_metric_spec(metric_spec)
    n_slices = len(triangle.slices)
    n_metrics = len(metric_spec)
    max_cols = ncols or _determine_facet_cols(n_slices * n_metrics)
    fig = _build_metric_slice_charts(
        _remove_triangle_samples(triangle) if hide_samples else triangle,
        plot_func=_plot_histogram,
        metric_dict=metric_dict,
        title=main_title,
        right_edge=right_edge,
        facet_titles=facet_titles,
        width=width,
        height=height,
        ncols=max_cols,
    ).configure_axis(**_compute_font_sizes(max_cols))
    return fig


def _plot_histogram(
    triangle: Triangle,
    metric: MetricFunc,
    name: str,
    right_edge: bool,
    title: alt.Title,
) -> alt.Chart:
    if right_edge:
        triangle = triangle.right_edge

    snake_name = _to_snake_case(name)

    metric_data = alt.Data(
        values=[
            {
                name: value,
                "iteration": i,
            }
            for cell in triangle
            for i, value in enumerate(_scalar_or_array_to_iter(metric(cell)))
        ]
    )
    data = alt.Data(values=build_plot_data(triangle, {name: metric}))

    histogram = (
        alt.Chart(data, title=title)
        .mark_bar()
        .encode(
            x=alt.X(f"{snake_name}.mean:Q", axis=alt.Axis(format=".2s"))
            .bin({"maxbins": 50})
            .title(name),
            y=alt.Y("count()").title("Count"),
        )
    )

    return histogram


def _build_metric_slice_charts(
    triangle, plot_func, title, facet_titles, width, height, ncols, **plot_kwargs
):
    charts = []
    n_slices = len(triangle.slices)
    for i, (_, triangle_slice) in enumerate(triangle.slices.items()):
        if facet_titles is None:
            slice_title = _slice_label(triangle_slice, triangle)
        else:
            slice_title = facet_titles[i]
        if plot_kwargs.get("metric_dict") is not None:
            for name, metric in plot_kwargs["metric_dict"].items():
                metric_title = (
                    (n_slices > 1) * (slice_title + ": ") + name
                    if facet_titles is None
                    else slice_title
                )
                charts.append(
                    plot_func(
                        triangle=triangle_slice,
                        metric=metric,
                        name=name,
                        title=alt.Title(metric_title, **SLICE_TITLE_KWARGS),
                        **{k: v for k, v in plot_kwargs.items() if k != "metric_dict"},
                    ).properties(width=width, height=height)
                )
        else:
            charts.append(
                plot_func(
                    triangle=triangle_slice,
                    title=alt.Title(slice_title, **SLICE_TITLE_KWARGS),
                    **plot_kwargs,
                ).properties(width=width, height=height)
            )
    fig = (
        _concat_charts(charts, title=title, ncols=ncols)
        .configure_axis(**_compute_font_sizes(ncols))
        .configure_legend(**_compute_font_sizes(ncols))
        .configure_mark(color="#1f8fff")
    )
    return fig


def _core_plot_data(cell: Cell) -> dict[str, Any]:
    return {
        "period_start": pd.to_datetime(cell.period_start),
        "period_end": pd.to_datetime(cell.period_end),
        "evaluation_date": pd.to_datetime(cell.evaluation_date),
        "dev_lag": cell.dev_lag(),
    }


def _calculate_field_summary(
    cell: Cell,
    prev_cell: Cell | None,
    func: MetricFunc,
    name: str,
    keep_samples: bool = False,
) -> FieldSummary:
    metric = _safe_apply_metric(cell, prev_cell, func)

    if metric is None or np.isscalar(metric) or len(metric) == 1:
        return FieldSummary(name, metric)

    return FieldSummary.from_metric(name, metric, keep_samples=keep_samples)


def _safe_apply_metric(cell: Cell, prev_cell: Cell | None, func: MetricFunc):
    try:
        if prev_cell.period != cell.period:
            raise IndexError
        return func(cell, prev_cell)
    except Exception:
        try:
            return func(cell)
        except Exception:
            return None


def _compute_font_sizes(mark_scaler: int) -> dict[str, float | int]:
    return {
        "titleFontSize": BASE_AXIS_TITLE_FONT_SIZE
        * np.exp(-FONT_SIZE_DECAY_FACTOR * (mark_scaler - 1)),
        "labelFontSize": BASE_AXIS_LABEL_FONT_SIZE
        * np.exp(-FONT_SIZE_DECAY_FACTOR * (mark_scaler - 1)),
    }


def _currency_symbol(triangle: Triangle) -> str:
    code = triangle.metadata[0].currency
    return get_currency_symbol(code, locale="en_US") or "$"


def _concat_charts(charts: list[alt.Chart], ncols: int, **kwargs) -> alt.Chart:
    if len(charts) == 1:
        return charts[0].properties(**kwargs)

    fig = alt.concat(*charts, columns=ncols, **kwargs)
    return fig


def _determine_facet_cols(n: int):
    """This is a replication of grDevices::n2mfrow in R"""
    return int(min(n, np.ceil(n / np.sqrt(n))))


def _slice_label(slice_tri: Triangle, base_tri: Triangle):
    slice_metadata = metadata_diff(base_tri.common_metadata, slice_tri.common_metadata)

    # Custom elements
    custom_elems = []
    for label, value in {
        **slice_metadata.details,
        **slice_metadata.loss_details,
    }.items():
        custom_elems.append(f"{string.capwords(label)}: {value}")

    # Bare elements
    bare_elems = []
    if slice_metadata.country is not None:
        bare_elems.append(slice_metadata.country)
    if slice_metadata.reinsurance_basis is not None:
        bare_elems.append(slice_metadata.reinsurance_basis)
    if slice_metadata.loss_definition is not None:
        bare_elems.append(slice_metadata.loss_definition)

    # Decorated elements
    decorated_elems = []
    if slice_metadata.per_occurrence_limit is not None:
        decorated_elems.append(f"limit {slice_metadata.per_occurrence_limit}")
    if slice_metadata.risk_basis is not None:
        decorated_elems.append(f"{slice_metadata.risk_basis} Basis")
    if slice_metadata.currency is not None:
        decorated_elems.append(f"in {slice_metadata.currency}")

    custom_label = ", ".join(custom_elems)
    bare_label = " ".join(bare_elems)
    decorated_label = "(" + ", ".join(decorated_elems) + ")"

    label = ""
    if custom_label:
        label += custom_label
    if label and bare_label:
        label += "; "
    if bare_label:
        label += bare_label
    if label and len(decorated_label) > 2:
        label += " "
    if len(decorated_label) > 2:
        label += decorated_label

    return label


def _scalar_or_array_to_iter(x: float | int | list | np.ndarray) -> np.ndarray:
    if np.isscalar(x) or x is None:
        return np.array([x])
    return x
